
from typing import *
from abc import abstractmethod
from SohuPrompt.data_utils.utils import InputExample

class DataProcessor:
    """
    labels of the dataset is optional

    here's the examples of loading the labels:

    :obj:`I`: ``DataProcessor(labels = ['positive', 'negative'])``

    :obj:`II`: ``DataProcessor(labels_path = 'datasets/labels.txt')``
    labels file should have label names separated by any blank characters, such as

    ..  code-block::

        positive neutral
        negative

    Args:
        labels (:obj:`Sequence[Any]`, optional): class labels of the dataset. Defaults to None.
        labels_path (:obj:`str`, optional): Defaults to None. If set and :obj:`labels` is None, load labels from :obj:`labels_path`.
    """

    def __init__(self,
                 labels: Optional[Sequence[Any]] = None,
                 labels_path: Optional[str] = None
                ):
        if labels is not None:
            self.labels = labels
        elif labels_path is not None:
            with open(labels_path, "r") as f:
                self.labels = ' '.join(f.readlines()).split()

    @property
    def labels(self) -> List[Any]:
        if not hasattr(self, "_labels"):
            raise ValueError("DataProcessor doesn't set labels or label_mapping yet")
        return self._labels

    @labels.setter
    def labels(self, labels: Sequence[Any]):
        if labels is not None:
            self._labels = labels
            self._label_mapping = {k: i for (i, k) in enumerate(labels)}

    @property
    def label_mapping(self) -> Dict[Any, int]:
        if not hasattr(self, "_labels"):
            raise ValueError("DataProcessor doesn't set labels or label_mapping yet")
        return self._label_mapping

    @label_mapping.setter
    def label_mapping(self, label_mapping: Mapping[Any, int]):
        self._labels = [item[0] for item in sorted(label_mapping.items(), key=lambda item: item[1])]
        self._label_mapping = label_mapping

    @property
    def id2label(self) -> Dict[int, Any]:
        if not hasattr(self, "_labels"):
            raise ValueError("DataProcessor doesn't set labels or label_mapping yet")
        return {i: k for (i, k) in enumerate(self._labels)}


    def get_label_id(self, label: Any) -> int:
        """get label id of the corresponding label

        Args:
            label: label in dataset

        Returns:
            int: the index of label
        """
        return self.label_mapping[label] if label is not None else None

    def get_labels(self) -> List[Any]:
        """get labels of the dataset

        Returns:
            List[Any]: labels of the dataset
        """
        return self.labels

    def get_num_labels(self):
        """get the number of labels in the dataset

        Returns:
            int: number of labels in the dataset
        """
        return len(self.labels)

    def get_train_examples(self, data_dir: Optional[str] = None) -> InputExample:
        """
        get train examples from the training file under :obj:`data_dir`

        call ``get_examples(data_dir, "train")``, see :py:meth:`~SohuPrompt.data_utils.data_processor.DataProcessor.get_examples`
        """
        return self.get_examples(data_dir, "train")

    def get_dev_examples(self, data_dir: Optional[str] = None) -> List[InputExample]:
        """
        get dev examples from the development file under :obj:`data_dir`

        call ``get_examples(data_dir, "dev")``, see :py:meth:`~SohuPrompt.data_utils.data_processor.DataProcessor.get_examples`
        """
        return self.get_examples(data_dir, "dev")

    def get_test_examples(self, data_dir: Optional[str] = None) -> List[InputExample]:
        """
        get test examples from the test file under :obj:`data_dir`

        call ``get_examples(data_dir, "test")``, see :py:meth:`~SohuPrompt.data_utils.data_processor.DataProcessor.get_examples`
        """
        return self.get_examples(data_dir, "test")

    def get_unlabeled_examples(self, data_dir: Optional[str] = None) -> List[InputExample]:
        """
        get unlabeled examples from the unlabeled file under :obj:`data_dir`

        call ``get_examples(data_dir, "unlabeled")``, see :py:meth:`~SohuPrompt.data_utils.data_processor.DataProcessor.get_examples`
        """
        return self.get_examples(data_dir, "unlabeled")

    @abstractmethod
    def get_examples(self, data_dir: Optional[str] = None, split: Optional[str] = None) -> List[InputExample]:
        """get the :obj:`split` of dataset under :obj:`data_dir`

        :obj:`data_dir` is the base path of the dataset, for example:

        training file could be located in ``data_dir/train.txt``

        Args:
            data_dir (str): the base path of the dataset
            split (str): ``train`` / ``dev`` / ``test`` / ``unlabeled``

        Returns:
            List[InputExample]: return a list of :py:class:`~SohuPrompt.data_utils.data_utils.InputExample`
        """
        raise NotImplementedError


from .processor import *

class ChineseSTS(CLSProcessor):
    """
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "0.0", "1.0", "2.0", "3.0", "4.0", "5.0",
            ],
            labels_mapped = [
                "完全没关系", "基本没关系", "没什么关系", "有点关系", "基本一样", "完全一样",
            ]
        )

    def get_examples(self, data_dir, split):
        path = os.path.join(data_dir, f"{split}.json")
        
        with open(path, encoding='utf8') as f:
            for line in f:
                example_json = json.loads(line)
                example = InputExample(
                    meta = {
                        "text_1": example_json["text_1"],
                        "text_2": example_json["text_2"],
                        "options": self.labels_mapped,
                    },
                    tgt_text = self.get_label(example_json["label"]),
                )
                examples.append(example)
                
        

    def get_templates(self):
        return [
            '文段一：{text_1} 文段二：{text_2} 问题：上述两个文段有什么关系？{options}',
        ]


class AFQMC(CLSProcessor):
    """
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "0", "1",
            ],
            labels_mapped = [
                "否", "是",
            ]
        )

    def get_examples(self, data_dir, split):
        path = os.path.join(data_dir, f"{split}.json")
        
        with open(path, encoding='utf8') as f:
            for line in f:
                example_json = json.loads(line)
                example = InputExample(
                    meta = {
                        "text_1": example_json["sentence1"],
                        "text_2": example_json["sentence2"],
                        "options": self.labels_mapped,
                    },
                    tgt_text = self.get_label(example_json["label"]) if "label" in example_json else None,
                )
                examples.append(example)
                
        

    def get_templates(self):
        return [
            '文段一：{text_1} 文段二：{text_2} 问题：上述两个文段表达的意思是否一致？{options}',
        ]



from .processor import *

class CMNLI(CLSProcessor):
    """
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "entailment", "contradiction", "neutral",
            ],
            labels_mapped = [
                "蕴含", "矛盾", "无关",
            ]
        )

    def get_examples(self, data_dir, split):
        path = os.path.join(data_dir, f"{split}.jsonl")
        examples = []
        with open(path, encoding='utf8') as f:
            for line_i, line in enumerate(f):
                # print(line_i)
                example_json = json.loads(line)
                example = InputExample(
                    meta = {
                        "premise": example_json["sentence1"],
                        "hypothesis": example_json["sentence2"],
                        "options": self.labels_mapped,
                    },
                    label = self.get_label(example_json["label"]),
                )
                examples.append(example)
        return examples
                
        

    def get_templates(self):
        return [
            '前提：{premise} 假设: {hypothesis} 问题：前提和假设是什么关系? {options}',
        ]

class CSNLI(CLSProcessor):
    """
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "entailment", "contradiction", "neutral",
            ],
            labels_mapped = [
                "蕴含", "矛盾", "无关",
            ]
        )

    def get_examples(self, data_dir, split):
        path = os.path.join(data_dir, f"{split}.jsonl")
        
        with open(path, encoding='utf8') as f:
            for line in f:
                example_json = json.loads(line)
                example = InputExample(
                    meta = {
                        "premise": example_json["sentence1"],
                        "hypothesis": example_json["sentence2"],
                        "options": self.labels_mapped,
                    },
                    tgt_text = self.get_label(example_json["label"]),
                )
                examples.append(example)
                
        

    def get_templates(self):
        return [
            '前提：{premise} 假设: {hypothesis} 问题：前提和假设是什么关系? {options}',
        ]

class OCNLI(CLSProcessor):
    """
    @inproceedings{ocnli,
	title={OCNLI: Original Chinese Natural Language Inference},
	author={Hai Hu and Kyle Richardson and Liang Xu and Lu Li and Sandra Kuebler and Larry Moss},
	booktitle={Findings of EMNLP},
	year={2020},
	url={https://arxiv.org/abs/2010.05444}
    }
    """

    def __init__(self):
        super().__init__(
            labels_origin = [
                "entailment", "contradiction", "neutral",
            ],
            labels_mapped = [
                "蕴含", "矛盾", "无关",
            ]
        )

    def get_examples(self, data_dir, split):
        path = os.path.join(data_dir, f"{split}.json")
        
        with open(path, encoding='utf8') as f:
            for line in f:
                example_json = json.loads(line)
                if "label" not in example_json or example_json["label"] not in self.labels_origin: continue
                example = InputExample(
                    meta = {
                        "premise": example_json["sentence1"],
                        "hypothesis": example_json["sentence2"],
                        "options": self.labels_mapped,
                    },
                    tgt_text = self.get_label(example_json["label"]),
                )
                examples.append(example)
                
        

    def get_templates(self):
        return [
            '前提：{premise} 假设: {hypothesis} 问题：前提和假设是什么关系? {options}',
        ]


from .processor import *

class Mandarinograd(CLSProcessor):
    """
    @inproceedings{bernard2020mandarinograd,
    title={Mandarinograd: A Chinese Collection of Winograd Schemas},
    author={Bernard, Timoth{\'e}e and Han, Ting},
    booktitle={Proceedings of The 12th Language Resources and Evaluation Conference},
    pages={21--26},
    year={2020}
    }
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "0", "1",
            ],
            labels_mapped = [
                "不可以", "可以", 
            ]
        )

    def get_examples(self, data_dir, split):
        path = os.path.join(data_dir, f"{split}.json")
        
        with open(path, encoding='utf8') as f:
            for example_json in json.load(f).values():
                example = InputExample(
                    meta = {
                        "before": example_json["text"],
                        "after": example_json["hypothesis"],
                        "options": self.labels_mapped,
                    },
                    tgt_text = self.get_label(example_json["class_id"]),
                )
                examples.append(example)
                
        

    def get_templates(self):
        return [
            '文本：{before} 问题：上述文本进行指代消解后，可以理解为 "{after}" 吗？{options}',
        ]


class CLUEWSC2020(CLSProcessor):
    """
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "false", "true",
            ],
            labels_mapped = [
                "否", "是",
            ]
        )

    def get_examples(self, data_dir, split):
        path = os.path.join(data_dir, f"{split}.json")
        
        with open(path, encoding='utf8') as f:
            for line in f:
                example_json = json.loads(line)
                example = InputExample(
                    meta = {
                        "that": example_json["target"]["span1_text"],
                        "it": example_json["target"]["span2_text"],
                        "text": example_json["text"],
                        "options": self.labels_mapped,
                    },
                    tgt_text = self.get_label(example_json["label"]),
                )
                examples.append(example)
                
        

    def get_templates(self):
        return [
            '文本：{text} 问题：在上文中，“{it}“ 指代 "{that}" 吗？{options}',
        ]


import itertools
import warnings
import torch
import torch.nn as nn
from torch.utils.data import Dataset
from typing import Union, List, Tuple, Dict, Optional
from transformers.data.processors.utils import InputExample, InputFeatures
from transformers.tokenization_utils import PreTrainedTokenizer
from collections import defaultdict
from SohuPrompt.utils import round_list
import numpy as np


class TokenizerWrapper:
    def __init__(
        self,
        max_seq_length: int,
        tokenizer: PreTrainedTokenizer,
        truncate_method: Optional[str] = "tail",
        create_token_type_ids: Optional[str] = False,
        **kwargs,
    ):
        self.max_seq_length = max_seq_length

        self.tokenizer = tokenizer
        if truncate_method == "tail":
            self.truncate_fct = self.truncate_from_tail
        elif truncate_method == "head":
            self.truncate_fct = self.truncate_from_head
        elif truncate_method == "balanced":
            self.truncate_fct = self.balanced_truncate
        else:
            raise NotImplementedError

        self.create_token_type_ids = create_token_type_ids

        self.template_mask_token = "<mask>"
        self.template_eos_token = "<eos>"
        self.template_bos_token = "<bos>"
        self.template_sep_token = "<sep>"
        self.template_cls_token = "<cls>"
        self.template_pad_token = "<pad>"

        from transformers import logging

        verbosity_before = logging.get_verbosity()
        logging.set_verbosity(logging.CRITICAL)  # TODO solve this in a more elegant way
        self.mask_token_map = {
            self.template_mask_token: self.tokenizer.mask_token
            if hasattr(self.tokenizer, "mask_token")
            else ""
        }
        self.eos_token_map = {
            self.template_eos_token: self.tokenizer.eos_token
            if hasattr(self.tokenizer, "eos_token")
            else ""
        }
        self.bos_token_map = {
            self.template_bos_token: self.tokenizer.bos_token
            if hasattr(self.tokenizer, "bos_token")
            else ""
        }
        self.sep_token_map = {
            self.template_sep_token: self.tokenizer.sep_token
            if hasattr(self.tokenizer, "sep_token")
            else ""
        }
        self.cls_token_map = {
            self.template_cls_token: self.tokenizer.cls_token
            if hasattr(self.tokenizer, "cls_token")
            else ""
        }
        self.pad_token_map = {
            self.template_pad_token: self.tokenizer.pad_token
            if hasattr(self.tokenizer, "pad_token")
            else ""
        }
        logging.set_verbosity(verbosity_before)

        self.num_truncated_sentences = 0
        self.total_passed_sentences = 0

    @property
    def truncate_rate(
        self,
    ):
        r"""Using this function, one can easily identify how many sentence has be truncated, thus help the user to choose a better thresthold for chunking."""
        if self.total_passed_sentences == 0:
            return None
        else:
            return self.num_truncated_sentences / self.total_passed_sentences

    @property
    def special_tokens_maps(
        self,
    ) -> Dict:
        r"""This need to be specified in specific language model"""
        if not hasattr(self, "_special_tokens_map"):
            _special_tokens_map = {}
            for attrname in self.__dict__.keys():
                if attrname.endswith("_token_map"):
                    _special_tokens_map.update(getattr(self, attrname))
        return _special_tokens_map

    def tokenize_with_mask(
        self,
        wrapped_example: List[Dict],
    ) -> InputFeatures:
        raise NotImplementedError

    def tokenize_without_mask(
        self,
        wrapped_example: List[Dict],
    ) -> InputFeatures:
        raise NotImplementedError

    @staticmethod
    def balanced_truncate(input_dict: Dict, num_tokens_to_truncate: int = 0) -> Dict:
        """truncate the inputs with balance, number of cut tokens is proportional to the part's length."""
        shortenable_lens = [
            len(parts) if parts[0] == 1 else 0
            for parts in input_dict["shortenable_ids"]
        ]
        total_shortenable_len = sum(shortenable_lens)
        num_tokens_to_truncate_each_part = [
            part_len / total_shortenable_len * num_tokens_to_truncate
            for part_len in shortenable_lens
        ]
        round_list(num_tokens_to_truncate_each_part, num_tokens_to_truncate)

        truncated_example = defaultdict(list)
        for key in input_dict:
            parts = input_dict[key]
            for num_tokens_to_truncate_part, part in zip(
                num_tokens_to_truncate_each_part, parts
            ):
                truncated_example[key].append(
                    part[: len(part) - num_tokens_to_truncate_part]
                )
        return truncated_example

    @staticmethod
    def truncate_from_tail(input_dict: Dict, num_tokens_to_truncate: int = 0) -> Dict:
        r"""truncate the inputs from the rear"""
        truncated_example = defaultdict(list)
        shortenable_ids = input_dict["shortenable_ids"]
        for key in input_dict:
            parts = input_dict[key]
            to_trunc = num_tokens_to_truncate
            for i, part in enumerate(parts[::-1]):
                if len(part) == 0:  # to prevent some part are empty after tokenization
                    continue
                if (
                    shortenable_ids[-1 - i][0] == 0
                ):  # ==0 means the part is not shortenable
                    continue
                parts[-1 - i] = part[:-to_trunc] if to_trunc < len(part) else []
                to_trunc -= len(part)
                if to_trunc <= 0:
                    break
            truncated_example[key] = parts
        return truncated_example

    @staticmethod
    def truncate_from_head(input_dict: Dict, num_tokens_to_truncate: int = 0) -> Dict:
        r"""truncate the inputs from the head"""
        truncated_example = defaultdict(list)
        shortenable_ids = input_dict["shortenable_ids"]
        for key in input_dict:
            parts = input_dict[key]
            to_trunc = num_tokens_to_truncate
            for i, part in enumerate(parts):
                print(i,part)
                if shortenable_ids[i][0] == 0:  # ==0 means the part is not shortenable
                    continue
                parts[i] = part[:-to_trunc] if to_trunc < len(part) else []
                to_trunc -= len(part)
                if to_trunc <= 0:
                    break
            truncated_example[key] = parts
        return truncated_example

    @staticmethod
    def concate_parts(input_dict: Dict) -> Dict:
        for key in input_dict:
            input_dict[key] = list(itertools.chain(*input_dict[key]))
        return input_dict

    @staticmethod
    def padding(
        input_dict: Dict,
        max_len: int,
        pad_id_for_inputs: int = 0,
        pad_id_for_others: int = 0,
    ) -> None:
        for key, value in input_dict.items():
            if len(input_dict[key]) > max_len:
                print(len(input_dict[key]), max_len)
                raise ValueError(
                    f"""Truncated seq length of '{key}' still greater than max length {max_len}."\
                    "One possible reason is that no enough shortenable parts in template. Try adding {{"shortenable": "True"}} property.
                """
                )
            if "input" in key:
                input_dict[key].extend([pad_id_for_inputs] * (max_len - len(value)))
            else:
                input_dict[key].extend([pad_id_for_others] * (max_len - len(value)))
        return input_dict

    def add_special_tokens(self, encoder_inputs):
        # add special tokens
        for key in encoder_inputs:
            if key == "input_ids":
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    encoder_inputs[
                        key
                    ] = self.tokenizer.build_inputs_with_special_tokens(
                        encoder_inputs[key]
                    )
            else:
                special_tokens_mask = np.array(
                    self.tokenizer.get_special_tokens_mask(encoder_inputs[key])
                )
                with_special_tokens = np.array(
                    self.tokenizer.build_inputs_with_special_tokens(encoder_inputs[key])
                )
                if key in ["soft_token_ids"]:  # TODO maybe more than this
                    encoder_inputs[key] = (
                        (1 - special_tokens_mask) * with_special_tokens
                    ).tolist()  # use 0 as special
                else:
                    encoder_inputs[key] = (
                        (1 - special_tokens_mask) * with_special_tokens
                        - special_tokens_mask * 100
                    ).tolist()  # use -100 as special
        return encoder_inputs

    def truncate(self, encoder_inputs, num_special_times):
        # print('encoder_inputs',encoder_inputs['input_ids'])
        total_tokens = sum([len(part) for part in encoder_inputs["input_ids"]])
        num_specials = self.num_special_tokens_to_add
        num_tokens_to_truncate = total_tokens - self.max_seq_length + num_specials * num_special_times
        self.total_passed_sentences += 1
        if num_tokens_to_truncate > 0:
            self.num_truncated_sentences += 1
            encoder_inputs = self.truncate_fct(
                input_dict=encoder_inputs, num_tokens_to_truncate=num_tokens_to_truncate
            )
        return encoder_inputs



import json
from transformers.tokenization_utils import PreTrainedTokenizer
from yacs.config import CfgNode
from SohuPrompt.data_utils import InputFeatures
import re
from SohuPrompt import Verbalizer
from typing import *
import torch
import torch.nn as nn
import torch.nn.functional as F
from SohuPrompt.utils.logging import logger



class One2oneVerbalizer(Verbalizer):
    r"""
    The basic manually defined verbalizer class, this class is inherited from the :obj:`Verbalizer` class.
    This class restrict the use of label words to one words per label. For a verbalzer with less constraints,
    please use Basic ManualVerbalizer.

    Args:
        tokenizer (:obj:`PreTrainedTokenizer`): The tokenizer of the current pre-trained model to point out the vocabulary.
        classes (:obj:`classes`): The classes (or labels) of the current task.
        num_classes (:obj:`int`): Optional. The number of classes of the verbalizer. Only one of `classes` and `num_classes` should be used.
        label_words (:obj:`Union[Sequence[str], Mapping[str, str]]`, optional): The label words that are projected by the labels.
        prefix (:obj:`str`, optional): The prefix string of the verbalizer. (used in PLMs like RoBERTa, which is sensitive to prefix space)
        multi_token_handler (:obj:`str`, optional): The handling strategy for multiple tokens produced by the tokenizer.
        post_log_softmax (:obj:`bool`, optional): Whether to apply log softmax post processing on label_logits. Default to True.
    """
    def __init__(self,
                 tokenizer: PreTrainedTokenizer,
                 num_classes: Optional[int] = None,
                 classes: Optional[List] = None,
                 label_words: Optional[Union[Sequence[str], Mapping[str, str]]] = None,
                 prefix: Optional[str] = " ",
                 multi_token_handler: Optional[str] = "first",
                 post_log_softmax: Optional[bool] = True,
                ):
        super().__init__(tokenizer=tokenizer, num_classes=num_classes, classes=classes)
        self.prefix = prefix
        self.multi_token_handler = multi_token_handler
        self.label_words = label_words
        self.post_log_softmax = post_log_softmax

    def on_label_words_set(self):
        super().on_label_words_set()
        self.label_words = self.add_prefix(self.label_words, self.prefix)
        self.generate_parameters()

    @staticmethod
    def add_prefix(label_words, prefix):
        r"""Add prefix to label words. For example, if a label words is in the middle of a template,
        the prefix should be ``' '``.

        Args:
            label_words (:obj:`Union[Sequence[str], Mapping[str, str]]`, optional): The label words that are projected by the labels.
            prefix (:obj:`str`, optional): The prefix string of the verbalizer.

        Returns:
            :obj:`Sequence[str]`: New label words with prefix.
        """
        new_label_words = []
        if isinstance(label_words[0], list):
            assert max([len(w) for w in label_words]) ==  1, "Providing multiple label words, you should use other verbalizers instead."
            label_words = [w[0] for w in label_words]

        for word in label_words:
            if word.startswith("<!>"):
                new_label_words.append(word.split("<!>")[1])
            else:
                new_label_words.append(prefix + word)

        return new_label_words

    def generate_parameters(self) -> List:
        r"""In basic manual template, the parameters are generated from label words directly.
        In this implementation, the label_words should not be tokenized into more than one token.
        """
        words_ids = []
        for word in self.label_words:
            word_ids = self.tokenizer.encode(word, add_special_tokens=False)
            if len(word_ids) > 1:
                logger.warning("Word {} is split into multiple tokens: {}. \
                    If this is not what you expect, try using another word for this verbalizer" \
                    .format(word, self.tokenizer.convert_ids_to_tokens(word_ids)))
            words_ids.append(word_ids)


        max_len  = max([len(ids) for ids in words_ids])
        words_ids_mask = [[1]*len(ids) + [0]*(max_len-len(ids)) for ids in words_ids]
        words_ids = [ids+[0]*(max_len-len(ids)) for ids in words_ids]

        words_ids_tensor = torch.tensor(words_ids)
        words_ids_mask = torch.tensor(words_ids_mask)
        self.label_words_ids = nn.Parameter(words_ids_tensor, requires_grad=False)
        self.label_words_mask = nn.Parameter(words_ids_mask, requires_grad=False)

    def project(self,
                logits: torch.Tensor,
                **kwargs,
                ) -> torch.Tensor:
        r"""
        Project the labels, the return value is the normalized (sum to 1) probs of label words.

        Args:
            logits (:obj:`torch.Tensor`): The original logits of label words.

        Returns:
            :obj:`torch.Tensor`: The normalized logits of label words
        """
        label_words_logits = logits[:, self.label_words_ids]
        label_words_logits = self.handle_multi_token(label_words_logits, self.label_words_mask)
        return label_words_logits

    def process_logits(self, logits: torch.Tensor, **kwargs):
        r"""A whole framework to process the original logits over the vocabulary, which contains four steps:

        (1) Project the logits into logits of label words

        if self.post_log_softmax is True:

            (2) Normalize over all label words

            (3) Calibrate (optional)

        Args:
            logits (:obj:`torch.Tensor`): The original logits.

        Returns:
            (:obj:`torch.Tensor`): The final processed logits over the label words set.
        """
        # project
        label_words_logits = self.project(logits, **kwargs)  #Output: (batch_size, num_classes) or  (batch_size, num_classes, num_label_words_per_label)

        if self.post_log_softmax:
            # normalize
            label_words_probs = self.normalize(label_words_logits)

            # calibrate
            if  hasattr(self, "_calibrate_logits") and self._calibrate_logits is not None:
                label_words_probs = self.calibrate(label_words_probs=label_words_probs)

            # convert to logits
            label_words_logits = torch.log(label_words_probs+1e-15)

        return label_words_logits

    def normalize(self, logits: torch.Tensor) -> torch.Tensor:
        """
        Given logits regarding the entire vocabulary, return the probs over the label words set.

        Args:
            logits (:obj:`Tensor`): The logits over the entire vocabulary.

        Returns:
            :obj:`Tensor`: The logits over the label words set.

        """
        batch_size = logits.shape[0]
        return F.softmax(logits.reshape(batch_size, -1), dim=-1).reshape(*logits.shape)


    def calibrate(self, label_words_probs: torch.Tensor, **kwargs) -> torch.Tensor:
        r"""

        Args:
            label_words_probs (:obj:`torch.Tensor`): The probability distribution of the label words with the shape of [``batch_size``, ``num_classes``, ``num_label_words_per_class``]

        Returns:
            :obj:`torch.Tensor`: The calibrated probability of label words.
        """
        shape = label_words_probs.shape
        assert self._calibrate_logits.dim() ==  1, "self._calibrate_logits are not 1-d tensor"
        calibrate_label_words_probs = self.normalize(self.project(self._calibrate_logits.unsqueeze(0), **kwargs))
        assert calibrate_label_words_probs.shape[1:] == label_words_probs.shape[1:] \
             and calibrate_label_words_probs.shape[0]==1, "shape not match"
        label_words_probs /= (calibrate_label_words_probs+1e-15)
        # normalize # TODO Test the performance
        norm = label_words_probs.reshape(shape[0], -1).sum(dim=-1,keepdim=True) # TODO Test the performance of detaching()
        label_words_probs /= norm
        return label_words_probs












"""
This file contains the logic for loading data for all typing tasks.
# TODO license
"""

import os
import json, csv
from abc import ABC, abstractmethod
from collections import defaultdict, Counter
from typing import *

from transformers.tokenization_utils import SPECIAL_TOKENS_MAP_FILE

from SohuPrompt.utils.logging import logger

from SohuPrompt.data_utils.utils import InputExample
from SohuPrompt.data_utils.data_processor import DataProcessor



class FewNERDProcessor(DataProcessor):
    """
    `Few-NERD <https://ningding97.github.io/fewnerd/>`_ a large-scale, fine-grained manually annotated named entity recognition dataset

    It was released together with `Few-NERD: Not Only a Few-shot NER Dataset (Ning Ding et al. 2021) <https://arxiv.org/pdf/2105.07464.pdf>`_

    Examples:

    ..  code-block:: python

        from SohuPrompt.data_utils.typing_dataset import PROCESSORS

        base_path = "datasets/Typing"

        dataset_name = "FewNERD"
        dataset_path = os.path.join(base_path, dataset_name)
        processor = PROCESSORS[dataset_name.lower()]()
        train_dataset = processor.get_train_examples(dataset_path)
        dev_dataset = processor.get_dev_examples(dataset_path)
        test_dataset = processor.get_test_examples(dataset_path)

        assert processor.get_num_labels() == 66
        assert processor.get_labels() == [
            "person-actor", "person-director", "person-artist/author", "person-athlete", "person-politician", "person-scholar", "person-soldier", "person-other",
            "organization-showorganization", "organization-religion", "organization-company", "organization-sportsteam", "organization-education", "organization-government/governmentagency", "organization-media/newspaper", "organization-politicalparty", "organization-sportsleague", "organization-other",
            "location-GPE", "location-road/railway/highway/transit", "location-bodiesofwater", "location-park", "location-mountain", "location-island", "location-other",
            "product-software", "product-food", "product-game", "product-ship", "product-train", "product-airplane", "product-car", "product-weapon", "product-other",
            "building-theater", "building-sportsfacility", "building-airport", "building-hospital", "building-library", "building-hotel", "building-restaurant", "building-other",
            "event-sportsevent", "event-attack/battle/war/militaryconflict", "event-disaster", "event-election", "event-protest", "event-other",
            "art-music", "art-writtenart", "art-film", "art-painting", "art-broadcastprogram", "art-other",
            "other-biologything", "other-chemicalthing", "other-livingthing", "other-astronomything", "other-god", "other-law", "other-award", "other-disease", "other-medical", "other-language", "other-currency", "other-educationaldegree",
        ]
        assert dev_dataset[0].text_a == "The final stage in the development of the Skyfox was the production of a model with tricycle landing gear to better cater for the pilot training market ."
        assert dev_dataset[0].meta["entity"] == "Skyfox"
        assert dev_dataset[0].label == 30
    """
    def __init__(self):
        super().__init__()
        self.labels = [
            "person-actor", "person-director", "person-artist/author", "person-athlete", "person-politician", "person-scholar", "person-soldier", "person-other",
            "organization-showorganization", "organization-religion", "organization-company", "organization-sportsteam", "organization-education", "organization-government/governmentagency", "organization-media/newspaper", "organization-politicalparty", "organization-sportsleague", "organization-other",
            "location-GPE", "location-road/railway/highway/transit", "location-bodiesofwater", "location-park", "location-mountain", "location-island", "location-other",
            "product-software", "product-food", "product-game", "product-ship", "product-train", "product-airplane", "product-car", "product-weapon", "product-other",
            "building-theater", "building-sportsfacility", "building-airport", "building-hospital", "building-library", "building-hotel", "building-restaurant", "building-other",
            "event-sportsevent", "event-attack/battle/war/militaryconflict", "event-disaster", "event-election", "event-protest", "event-other",
            "art-music", "art-writtenart", "art-film", "art-painting", "art-broadcastprogram", "art-other",
            "other-biologything", "other-chemicalthing", "other-livingthing", "other-astronomything", "other-god", "other-law", "other-award", "other-disease", "other-medical", "other-language", "other-currency", "other-educationaldegree",
        ]

    def get_examples(self, data_dir, split):
        path = os.path.join(data_dir, "supervised/{}.txt".format(split))
        with open(path, encoding='utf8') as f:
            data = FewNERDProcessor.load_data(f)

            examples = []

            for idx, (xs, ys, spans) in enumerate(data):
                for span in spans:
                    text_a = " ".join(xs)
                    meta = {
                        "entity": " ".join(xs[span[0]: span[1]+1])
                    }
                    example = InputExample(guid=str(idx), text_a=text_a, meta=meta, label=self.get_label_id(ys[span[0]][2:]))
                    examples.append(example)

            return examples

    @staticmethod
    def load_data(file):
        data = []
        xs = []
        ys = []
        spans = []

        for line in file.readlines():
            pair = line.split()
            if pair == []:
                if xs != []:
                    data.append((xs, ys, spans))
                xs = []
                ys = []
                spans = []
            else:
                xs.append(pair[0])

                tag = pair[-1]
                if tag != 'O':
                    if len(ys) == 0 or tag != ys[-1][2:]:
                        tag = 'B-' + tag
                        spans.append([len(ys), len(ys)])
                    else:
                        tag = 'I-' + tag
                        spans[-1][-1] = len(ys)
                ys.append(tag)
        return data

PROCESSORS = {
    "fewnerd": FewNERDProcessor,
    # "conll2003": Conll2003Processor,
    # "ontonotes5_0": OntoNotes5_0Processor,
    # "bbn": BBNProcessor,
}



from SohuPrompt.data_utils.data_processor import *
import os, json, csv

class CLSProcessor(DataProcessor):
    def __init__(self, labels_origin, labels_mapped):
        self.labels_origin = labels_origin
        self.labels_mapped = labels_mapped
        self.labels_mapping = {
            origin: kth for kth, (origin, mapped) in enumerate(zip(labels_origin, labels_mapped))
        }

    def get_label(self, origin):
        return self.labels_mapping[origin]


from .processor import *

class ChnSentiCorp(CLSProcessor):
    """
    @inproceedings{st2008ChnSentiCorp,
    title={An empirical study of sentiment analysis for chinese documents},
    booktitle={Expert Systems with Applications},
    pages={2612--2619}
    author={Songbo Tan, Jin Zhang},
    year={2008}
    }
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "0", "1",
            ],
            labels_mapped = [
                "消极", "积极",
            ]
        )

    def get_examples(self, data_dir, split):
        path = os.path.join(data_dir, f"{split}.jsonl")
        
        examples = []
        with open(path, encoding='utf8') as f:
            for line in f:
                example_json = json.loads(line)
                example = InputExample(
                    meta = {
                        "context": example_json["text_a"],
                        "options": self.labels_mapped,
                    },
                    label = self.get_label(example_json["label"]),
                )
                examples.append(example)
        return examples
                
        

    def get_templates(self):
        return [
            '文本：{context} 问题:上述文本所表达的情感为？{options}',
        ]


class ECISA(CLSProcessor):
    """
    @article{徐琳宏2008情感词汇本体的构造,
    title={情感词汇本体的构造},
    author={徐琳宏 and 林鸿飞 and 潘宇 and 任惠 and 陈建美},
    journal={情报学报},
    volume={27},
    number={2},
    pages={180--185},
    year={2008}
    }
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "0", "1", "2",
            ],
            labels_mapped = [
                "不含情感", "褒义", "贬义",
            ]
        )

    def get_examples(self, data_dir, split):
        if split == 'dev': return []
        path = os.path.join(data_dir, f"{split}.json")
        
        with open(path, encoding='utf8') as f:
            for example_json in json.load(f):
                sents = example_json["Sentence"]
                if isinstance(sents, dict): sents = [sents]
                for i, sent in enumerate(sents):
                    if "label" in sent:
                        example = InputExample(
                            meta = {
                                "context_before": "".join([
                                    s["text"] for s in sents[:i]
                                ]),
                                "text": sent["text"],
                                "context_after": "".join([
                                    s["text"] for s in sents[i+1:]
                                ]),
                                "options": self.labels_mapped,
                            },
                            tgt_text = self.get_label(sent["label"]),
                        )
                        examples.append(example)
                
        
        
    def get_templates(self):
        return [
            '文本：{context_before} {text} {context_after} 问题：上述文本中，"{text}"所表达的情感为？{options}',
        ]


class JD_FULL(CLSProcessor):
    """
    @article{zx2017JDFull,
    author    = {Xiang Zhang and Yann LeCun},
    title     = {Which Encoding is the Best for Text Classification in Chinese, English, Japanese and Korean?},
    journal   = {CoRR},
    volume    = {abs/1708.02657},
    year      = {2017},
    url       = {http://arxiv.org/abs/1708.02657},
    archivePrefix = {arXiv},
    eprint    = {1708.02657}
    }
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "1", "2", "3", "4", "5",
            ],
            labels_mapped = [
                "1", "2", "3", "4", "5",
                # "很差 ", "差", "中", "好", "很好",
            ]
        )

    def get_examples(self, data_dir, split):
        if split == 'dev': return []
        path = os.path.join(data_dir, f"{split}.jsonl")
        
        with open(path, encoding='utf8') as f:
            for line in f:
                example_json = json.loads(line)
                example = InputExample(
                    meta = {
                        "title": example_json["title"],
                        "review": example_json["review"],
                        "options": self.labels_mapped,
                    },
                    tgt_text = self.get_label(example_json["label"]),
                )
                examples.append(example)
                
        
        
    def get_templates(self):
        return [
            '评价：{title} {review} 问题：据此分析，这段评价给出的评分为？{options}'
        ]


class SimplifyWeibo4Moods(CLSProcessor):
    """
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "0", "1", "2", "3"
            ],
            labels_mapped = [
                "喜悦", "愤怒", "厌恶", "低落"
            ]
        )

    def get_examples(self, data_dir, split):
        if split!='train': return []
        path = os.path.join(data_dir, f"{split}.csv")
        
        with open(path, encoding='utf8') as f:
            reader = csv.reader(f, delimiter=",")
            for i, row in enumerate(reader):
                if i==0: continue
                label, review = row[0], ",".join(row[1:])
                example = InputExample(
                    meta = {
                        "context": review,
                        "options": self.labels_mapped,
                    },
                    tgt_text = self.get_label(label),
                )
                examples.append(example)
                
        

    def get_templates(self):
        return [
            '文本：{context} 问题：上述文本所表达的情感为？{options}',
        ]


class PositiveNegative(CLSProcessor):
    """
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "0", "1",
            ],
            labels_mapped = [
                "消极", "积极",
            ]
        )

    def get_examples(self, data_dir, split):
        if split != 'train': raise ValueError
        path = os.path.join(data_dir, f"{split}.csv")
        
        with open(path, encoding='utf8') as f:
            reader = csv.reader(f, delimiter=",")
            for i, row in enumerate(reader):
                if i==0: continue
                label, review = row[0], ",".join(row[1:])
                example = InputExample(
                    meta = {
                        "context": review,
                        "options": self.labels_mapped,
                    },
                    tgt_text = self.get_label(label),
                )
                examples.append(example)
                
        

    def get_templates(self):
        return [
            '文本：{context} 问题：上述文本所表达的情感为？{options}',
        ]


class RatingMovie(CLSProcessor):
    """
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "1", "2", "3", "4", "5",
            ],
            labels_mapped = [
                "1", "2", "3", "4", "5",
            ]
        )

    def get_examples(self, data_dir, split):
        if split != 'train': raise ValueError
        path = os.path.join(data_dir, f"{split}.csv")
        
        with open(path, encoding='utf8') as f:
            reader = csv.reader(f, delimiter=",")
            for i, row in enumerate(reader):
                if i==0: continue
                rating, text = row[2], row[4] 
                example = InputExample(
                    meta = {
                        "text": text,
                        "options": self.labels_mapped,
                    },
                    tgt_text = self.get_label(rating),
                )
                examples.append(example)
                
        
        
    def get_templates(self):
        return [
            '评价：{text} 问题：据此估计，这段对电影的评价对应的评分为？{options}'
        ]


class RatingShopping(CLSProcessor):
    """
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "1", "2", "3", "4", "5",
            ],
            labels_mapped = [
                "1", "2", "3", "4", "5",
            ]
        )

    def get_examples(self, data_dir, split):
        if split != 'train': raise ValueError
        path = os.path.join(data_dir, f"{split}.csv")
        
        with open(path, encoding='utf8') as f:
            reader = csv.reader(f, delimiter=",")
            for i, row in enumerate(reader):
                if i==0: continue
                rating, text = row[2], row[4]+","+row[5]
                if rating not in self.labels_origin: continue # illegal
                example = InputExample(
                    meta = {
                        "text": text,
                        "options": self.labels_mapped,
                    },
                    tgt_text = self.get_label(rating),
                )
                examples.append(example)
                
        
        
    def get_templates(self):
        return [
            '评价：{text} 问题：这段评价对商品的评分为？{options}'
        ]


class RatingDianping(CLSProcessor):
    """
    """
    def __init__(self):
        super().__init__(
            labels_origin = [
                "1.0", "2.0", "3.0", "4.0", "5.0",
            ],
            labels_mapped = [
                "1", "2", "3", "4", "5",
            ]
        )

    def get_examples(self, data_dir, split):
        if split != 'train': raise ValueError
        path = os.path.join(data_dir, f"{split}.csv")
        
        with open(path, encoding='utf8') as f:
            reader = csv.reader(f, delimiter=",")
            for i, row in enumerate(reader):
                if i==0: continue
                rating_env, rating_flavor, rating_service, text = row[3], row[4], row[5], row[7]
                if rating_env=="": continue # illegal data
                example = InputExample(
                    meta = {
                        "text": text,
                        "question": "环境",
                        "options": self.labels_mapped,
                    },
                    tgt_text = self.get_label(rating_env)
                )
                examples.append(example)
                example = InputExample(
                    meta = {
                        "text": text,
                        "question": "特色",
                        "options": self.labels_mapped,
                    },
                    tgt_text = self.get_label(rating_flavor)
                )
                examples.append(example)
                example = InputExample(
                    meta = {
                        "text": text,
                        "question": "设施",
                        "options": self.labels_mapped,
                    },
                    tgt_text = self.get_label(rating_service)
                )
                examples.append(example)
        
        
    def get_templates(self):
        return [
            '评价:{text} 问题：据此估计，这段评价对{question}的评分为？{options}'
        ]


from sklearn.metrics import f1_score, precision_score, recall_score, accuracy_score
from typing import *
from SohuPrompt.utils.logging import logger

def f1(p, r):
    if p == 0. or r == 0.:
        return 0.
    return 2*p*r/(p+r)

def label_path(label, label_path_sep):
    label = label.strip(label_path_sep)
    label_path = label.split(label_path_sep)
    label_set = []
    for i in range(len(label_path)):
        label_set.append(label_path_sep.join(label_path[:i+1]))
    return label_set

def loose_micro(labels, preds, id2label, label_path_sep):
    if id2label is None:
        raise ValueError("no id2label dict provided, cannot calculate loose_micro_f1 !")
    labels = [label_path(id2label[i], label_path_sep) for i in labels]
    preds = [label_path(id2label[i], label_path_sep) for i in preds]
    cnt_pred = 0
    cnt_label = 0
    cnt_correct = 0
    for label, pred in zip(labels, preds):
        label = set(label)
        pred = set(pred)
        cnt_pred += len(pred)
        cnt_label += len(label)
        cnt_correct += len(label.intersection(pred))
    p = cnt_correct/cnt_pred
    r = cnt_correct/cnt_label
    f = f1(p, r)
    return {'precision': p, 'recall': r, 'f1': f}

def loose_macro(labels, preds, id2label, label_path_sep):
    if id2label is None:
        raise ValueError("no id2label dict provided, cannot calculate loose_micro_f1 !")
    labels = [label_path(id2label[i], label_path_sep) for i in labels]
    preds = [label_path(id2label[i], label_path_sep) for i in preds]
    p = 0.
    r = 0.
    for label, pred in zip(labels, preds):
        label = set(label)
        pred = set(pred)
        if len(pred) > 0:
            p += len(label.intersection(pred))/len(pred)
        if len(label) > 0:
            r += len(label.intersection(pred))/len(label)
    p /= len(labels)
    r /= len(labels)
    f = f1(p, r)
    return {'precision': p, 'recall': r, 'f1': f}


def classification_metrics(preds: Sequence[int],
                           labels: Sequence[int],
                           metric: Optional[str] = "micro-f1",
                           id2label: Optional[Dict] = None,
                           label_path_sep: Optional[str] = '-',
                          ) -> float:
    """evaluation metrics for classification task.

    Args:
        preds (Sequence[int]): predicted label ids for each examples
        labels (Sequence[int]): gold label ids for each examples
        metric (str, optional): type of evaluation function, support 'micro-f1', 'macro-f1', 'accuracy', 'precision', 'recall'. Defaults to "micro-f1".

    Returns:
        score (float): evaluation score
    """
    
    if metric == "micro-f1":
        score = f1_score(labels, preds, average='micro')
    elif metric == "macro-f1":
        score = f1_score(labels, preds, average='macro')
    elif metric == "accuracy":
        score = accuracy_score(labels, preds)
    elif metric == "precision":
        score = precision_score(labels, preds)
    elif metric == "recall":
        score = recall_score(labels, preds)
    # only hierarchical label loose metric is supported TODO naive multilabel ?
    elif metric == 'loose-micro-f1': 
        score = loose_micro(labels, preds, id2label=id2label, label_path_sep=label_path_sep)['f1']
    elif metric == 'loose-macro-f1':
        score = loose_macro(labels, preds, id2label=id2label, label_path_sep=label_path_sep)['f1']
    elif metric == 'loose-micro-precision': 
        score = loose_micro(labels, preds, id2label=id2label, label_path_sep=label_path_sep)['precision']
    elif metric == 'loose-macro-precision':
        score = loose_macro(labels, preds, id2label=id2label, label_path_sep=label_path_sep)['precision']
    elif metric == 'loose-micro-recall': 
        score = loose_micro(labels, preds, id2label=id2label, label_path_sep=label_path_sep)['recall']
    elif metric == 'loose-macro-recall':
        score = loose_macro(labels, preds, id2label=id2label, label_path_sep=label_path_sep)['recall']
    else:
        raise ValueError("'{}' is not a valid evaluation type".format(metric))
    return score

def generation_metric(hypos,
                      refs, 
                      metric: Optional[str] = "sentence_bleu"):
    r"""Some basic metric function for generation. However, many generation tasks
    has their own evaluation bash scripts.

    Args:
        hypos (:obj:`str`) : the generated sentence.
        refs (:obj:`list(str)`) : the referenced (ground-truth) sentence.
        metric (:obj:`str`, `optional`) : the type of metric option

    Returns:
        score (float): evaluate score
    """
    if metric == "sentence_bleu":
        # a simple criterion to visualize the performance, not rigorous.
        import nltk
        try:
            nltk_path = str(nltk.data.find("tokenizers/punkt"))
            logger.info(f"using nltk from: {nltk_path}")
        except LookupError:
            nltk.download('punkt')

        from nltk.translate.bleu_score import sentence_bleu
        from nltk.tokenize import word_tokenize
        from nltk.translate.bleu_score import SmoothingFunction
        smoothie = SmoothingFunction().method4 # a function for smooth
        scores = []
        
        for ref, hypo in zip(refs, hypos):
            tokenized_rs = []
            ref = ref.split("\n")
            for r in ref:
                tokenized_rs.append(word_tokenize(r))
            hypo = word_tokenize(hypo)
            try:
                sc = sentence_bleu(tokenized_rs, hypo, smoothing_function=smoothie)
            except ValueError: # TODO ZeroDivisionError
                logger.warning("math domain error in bleu, set to 0.0. generated sentence: {}".format(hypo))
                sc = 0.0
            scores.append(sc)
        score = sum(scores)/len(scores)
        return score
    else:
        raise ValueError("'{}' is not a valid metric type.".format(metric))



from json import decoder

from transformers.tokenization_utils import PreTrainedTokenizer
from SohuPrompt.plms.utils import TokenizerWrapper
from typing import List, Dict, Optional
from collections import defaultdict
from SohuPrompt.utils.logging import logger

class T5TokenizerWrapper(TokenizerWrapper):
    r"""
    Given wrapped example, e.g. A fun movie ! it is <mask> and <mask> .
    The input tokens is:  A fun movie ! it is <extra_id_0> and <extra_id_1> . </s>
    The decoder tokens is: <pad> <extra_id_0> <extra_id_1> </s>
    The expected output is <extra_id_0> good interesting </s>
    """
    def __init__(self,
                 max_seq_length: int,
                 tokenizer: PreTrainedTokenizer,
                 truncate_method: Optional[str] = 'tail',
                 decoder_max_length: Optional[int] = 128,
                 decode_from_pad: Optional[bool] = True,
                 predict_eos_token: Optional[bool] = False,
                 **kwargs):
        super().__init__(max_seq_length=max_seq_length, tokenizer=tokenizer,truncate_method=truncate_method)
        self.decoder_max_length = decoder_max_length
        self.decode_from_pad = decode_from_pad
        self.predict_eos = predict_eos_token
        if self.create_token_type_ids:
            logger.warning("token_type_ids is not valid in T5. will be depreciated.")

    def mask_token(self,i):
        return self.tokenizer.additional_special_tokens[i]


    def mask_token_ids(self, i ):
        return self.tokenizer.additional_special_tokens_ids[i]

    @property
    def num_special_tokens_to_add(self):
        if not hasattr(self, '_num_specials'):
            self._num_specials = self.tokenizer.num_special_tokens_to_add()
        return self._num_specials


    def tokenize_one_example(self, wrapped_example, teacher_forcing):
        ''' # TODO doesn't consider the situation that input has two parts
        '''
        wrapped_example, others = wrapped_example

        if teacher_forcing:
            tgt_text = others['tgt_text']
            if isinstance(tgt_text, str):
                tgt_text = [tgt_text]


        encoder_inputs = defaultdict(list)

        num_mask_token_used = 0

        decoder_input_ids = []
        loss_ids =[]

        for piece_id, piece in enumerate(wrapped_example):
            if piece['text'] == self.template_mask_token:
                if teacher_forcing:
                    decoder_input_ids.append(self.mask_token_ids(num_mask_token_used))
                    if num_mask_token_used > 0:
                        # if used in multiple mask setting, the <extra_id_1> etc. are also required
                        # to be predicted.
                        loss_ids.append(1)
                    else:
                        loss_ids.append(0)
                    encode_text = [self.mask_token_ids(num_mask_token_used)]
                    tgt_text_ids = self.tokenizer.encode(" " + tgt_text[num_mask_token_used], add_special_tokens=False)
                    decoder_input_ids.extend(tgt_text_ids)
                    loss_ids.extend([1] * len(tgt_text_ids))
                else:
                    decoder_input_ids.append(self.mask_token_ids(num_mask_token_used))
                    #decoder_input_ids.append(self.tokenizer.pad_token_id)
                    encode_text = [self.mask_token_ids(num_mask_token_used)]
                    #encode_text = [self.tokenizer.pad_token_id]
                    loss_ids.append(1)
                num_mask_token_used += 1
            else:
                if piece['text'] in self.special_tokens_maps.keys():
                    to_replace = self.special_tokens_maps[piece['text']]
                    if to_replace is not None:
                        piece['text'] = to_replace
                    else:
                        raise KeyError("This tokenizer doesn't specify {} token.".format(piece['text']))

                if 'soft_token_ids' in piece and piece['soft_token_ids']!=0:
                    encode_text =  [0] # can be replace by any token, since these token will use their own embeddings
                else:
                    encode_text = self.tokenizer.encode(piece['text'], add_special_tokens=False)

            encoding_length = len(encode_text)

            encoder_inputs['input_ids'].append(encode_text)
            for key in piece:
                if key not in ['text', 'loss_ids']:
                    encoder_inputs[key].append([piece[key]]*encoding_length)

        # decoder input ids
        decoder_inputs = {'decoder_input_ids': decoder_input_ids, 'loss_ids':loss_ids}
        decoder_inputs = self.truncate_decoder_inputs(decoder_inputs)

        encoder_inputs = self.truncate(encoder_inputs=encoder_inputs, num_special_times=1)

        # delete shortenable ids
        encoder_inputs.pop("shortenable_ids")
        encoder_inputs = self.concate_parts(input_dict=encoder_inputs)
        encoder_inputs = self.add_special_tokens(encoder_inputs=encoder_inputs)

        # create special input ids
        encoder_inputs['attention_mask'] = [1] *len(encoder_inputs['input_ids'])
        # padding
        encoder_inputs = self.padding(input_dict=encoder_inputs, max_len=self.max_seq_length, pad_id_for_inputs=self.tokenizer.pad_token_id)

        all_input_ids = {**encoder_inputs, **decoder_inputs}
        return all_input_ids

    def truncate_decoder_inputs(self, inputs):
        if self.decode_from_pad:
            inputs['decoder_input_ids'].insert(0, self.tokenizer.pad_token_id)
            inputs['loss_ids'].insert(0, 0)

        for key in inputs:
            inputs[key] = inputs[key][:self.decoder_max_length - 1]

        if self.predict_eos:
            inputs['decoder_input_ids'].append(self.tokenizer.eos_token_id)
            inputs['loss_ids'].append(1)
        inputs = self.padding(inputs, max_len = self.decoder_max_length, pad_id_for_inputs=self.tokenizer.pad_token_id)
        return inputs

class T5LMTokenizerWrapper(TokenizerWrapper):
    r"""
    The tokenizerwrapper is for the t5-lm-adapted version proposed by
    `The Power of Scale for Parameter-Efficient Prompt Tuning <https://arxiv.org/abs/2104.08691>`_
    Since this model is a autogressive language model fashion, it only support generation from the
    and of the text.

    Given wrapped example, e.g. A fun movie ! it is {"mask"}
    The encoder input is :  A fun movie ! it is </s>
    (Note that </s> is added in T5 encoder inputs, this will yield better result compared to not using </s>)
    The decoder input is : <pad> <extra_id_0> </s>
    The expected output is : good
    Under teacher forcing mode, the decoder input is   <pad> <extra_id_0>  good </s>, where good and </s> requires loss
    """
    def __init__(self,
                 max_seq_length: int,
                 tokenizer: PreTrainedTokenizer,
                 truncate_method: Optional[str] = 'tail',
                 decoder_max_length: Optional[int] = 128,
                 decode_from_pad: Optional[bool] = True,
                 predict_eos_token: Optional[bool] = False,
                 **kwargs):
        super().__init__(max_seq_length=max_seq_length, tokenizer=tokenizer,truncate_method=truncate_method)
        self.decoder_max_length = decoder_max_length
        self.decode_from_pad = decode_from_pad
        self.predict_eos = predict_eos_token
        if self.create_token_type_ids:
            logger.warning("token_type_ids is not valid in T5. will be depreciated.")

    def mask_token(self,i):
        return self.tokenizer.additional_special_tokens[i]


    def mask_token_ids(self, i ):
        return self.tokenizer.additional_special_tokens_ids[i]

    @property
    def num_special_tokens_to_add(self):
        if not hasattr(self, '_num_specials'):
            self._num_specials = self.tokenizer.num_special_tokens_to_add()
        return self._num_specials


    def tokenize_one_example(self, wrapped_example, teacher_forcing):
        wrapped_example, others = wrapped_example
        if teacher_forcing:
            tgt_text = others['tgt_text']
            if isinstance(tgt_text, str):
                tgt_text = [tgt_text]

        encoder_inputs = defaultdict(list)

        num_mask_token_used = 0

        decoder_input_ids = []
        loss_ids =[]

        for piece_id, piece in enumerate(wrapped_example):
            if piece['text'] == self.template_mask_token:
                if teacher_forcing:
                    decoder_input_ids.append(self.mask_token_ids(num_mask_token_used))
                    loss_ids.append(0)
                    encode_text = []
                    tgt_text_ids = self.tokenizer.encode(" " + tgt_text[num_mask_token_used], add_special_tokens=False)
                    decoder_input_ids.extend(tgt_text_ids)
                    loss_ids.extend([1] * len(tgt_text_ids))
                else:
                    decoder_input_ids.append(self.mask_token_ids(num_mask_token_used))
                    encode_text = [] # not add extra_id_0 to input_ids
                    loss_ids.append(1)
                break
            else:
                if piece['text'] in self.special_tokens_maps.keys():
                    to_replace = self.special_tokens_maps[piece['text']]
                    if to_replace is not None:
                        piece['text'] = to_replace
                    else:
                        raise KeyError("This tokenizer doesn't specify {} token.".format(piece['text']))

                if 'soft_token_ids' in piece and piece['soft_token_ids']!=0:
                    encode_text =  [0] # can be replace by any token, since these token will use their own embeddings
                else:
                    encode_text = self.tokenizer.encode(piece['text'], add_special_tokens=False)

            encoding_length = len(encode_text)

            encoder_inputs['input_ids'].append(encode_text)
            for key in piece:
                if key not in ['text', 'loss_ids']:
                    encoder_inputs[key].append([piece[key]]*encoding_length)

        # decoder input ids
        decoder_inputs = {'decoder_input_ids': decoder_input_ids, 'loss_ids':loss_ids}
        decoder_inputs = self.truncate_decoder_inputs(decoder_inputs)

        encoder_inputs = self.truncate(encoder_inputs=encoder_inputs, num_special_times=1)

        # delete shortenable ids
        encoder_inputs.pop("shortenable_ids")
        encoder_inputs = self.concate_parts(input_dict=encoder_inputs)
        encoder_inputs = self.add_special_tokens(encoder_inputs=encoder_inputs)

        # create special input ids
        encoder_inputs['attention_mask'] = [1] *len(encoder_inputs['input_ids'])
        # padding
        encoder_inputs = self.padding(input_dict=encoder_inputs, max_len=self.max_seq_length, pad_id_for_inputs=self.tokenizer.pad_token_id)

        all_input_ids = {**encoder_inputs, **decoder_inputs}
        return all_input_ids

    def truncate_decoder_inputs(self, inputs):
        if self.decode_from_pad:
            inputs['decoder_input_ids'].insert(0, self.tokenizer.pad_token_id)
            inputs['loss_ids'].insert(0, 0)

        for key in inputs:
            inputs[key] = inputs[key][:self.decoder_max_length - 1]

        inputs['decoder_input_ids'].append(self.tokenizer.eos_token_id)
        if self.predict_eos:
            inputs['loss_ids'].append(1)
        else:
            inputs['loss_ids'].append(0)
        inputs = self.padding(inputs, max_len = self.decoder_max_length)
        return inputs




class CPM2TokenizerWrapper(TokenizerWrapper):
    r"""
    Given wrapped example, e.g. A fun movie ! it is <mask> and <mask> .
    The input tokens is:  A fun movie ! it is <s_0> and <s_1> . <eod>
    The decoder tokens is: <s> <s_0> <s_1> <eod>
    The expected output is: <s_0> good <s_1> interesting </s>
    Another solution is add verbalizer label words into the labels, thus generating
    'interesting' given 'good'. However, it is not implemented in this version.
    """
    def __init__(self,
                 max_seq_length: int,
                 tokenizer: PreTrainedTokenizer,
                 truncate_method: Optional[str] = 'tail',
                 decoder_max_length: Optional[int] = 128,
                 decode_from_start: Optional[bool] = True,
                 predict_eos_token: Optional[bool] = False,
                 **kwargs):
        super().__init__(max_seq_length=max_seq_length, tokenizer=tokenizer,truncate_method=truncate_method)
        self.decoder_max_length = decoder_max_length
        self.decode_from_start = decode_from_start
        self.predict_eos = predict_eos_token
        if self.create_token_type_ids:
            logger.warning("token_type_ids is not valid in T5. will be depreciated.")

    def mask_token(self,i):
        return self.tokenizer.additional_special_tokens[i]


    def mask_token_ids(self, i ):
        return self.tokenizer.additional_special_tokens_ids[i]

    @property
    def num_special_tokens_to_add(self):
        if not hasattr(self, '_num_specials'):
            self._num_specials = self.tokenizer.num_special_tokens_to_add()
        return self._num_specials


    def tokenize_one_example(self, wrapped_example, teacher_forcing):
        ''' # TODO doesn't consider the situation that input has two parts
        '''
        wrapped_example, others = wrapped_example

        if teacher_forcing:
            tgt_text = others['tgt_text']
            if isinstance(tgt_text, str):
                tgt_text = [tgt_text]

        encoder_inputs = defaultdict(list)

        num_mask_token_used = 0

        decoder_input_ids = []
        loss_ids =[0]

        for piece_id, piece in enumerate(wrapped_example):
            if piece['text'] == self.template_mask_token:
                if teacher_forcing:
                    decoder_input_ids.append(self.mask_token_ids(num_mask_token_used))
                    encode_text = [self.mask_token_ids(num_mask_token_used)]
                    tgt_text_ids = self.tokenizer.encode(" " + tgt_text[num_mask_token_used], add_special_tokens=False)
                    decoder_input_ids.extend(tgt_text_ids)
                    loss_ids.extend([1] * len(tgt_text_ids))
                    # decoder_input_ids.append(self.mask_token_ids(num_mask_token_used+1))
                    loss_ids.append(1)
                else:
                    decoder_input_ids.append(self.mask_token_ids(num_mask_token_used))
                    encode_text = [self.mask_token_ids(num_mask_token_used)]
                    # decoder_input_ids.append(self.mask_token_ids(num_mask_token_used+1))
                    loss_ids[-1] = 1 # shift loss_ids
                    loss_ids.append(0)
                num_mask_token_used += 1
            else:
                if piece['text'] in self.special_tokens_maps.keys():
                    to_replace = self.special_tokens_maps[piece['text']]
                    if to_replace is not None:
                        piece['text'] = to_replace
                    else:
                        raise KeyError("This tokenizer doesn't specify {} token.".format(piece['text']))

                if 'soft_token_ids' in piece and piece['soft_token_ids']!=0:
                    encode_text =  [0] # can be replace by any token, since these token will use their own embeddings
                else:
                    encode_text = self.tokenizer.encode(piece['text'], add_special_tokens=False)

            encoding_length = len(encode_text)

            encoder_inputs['input_ids'].append(encode_text)
            for key in piece:
                if key not in ['text', 'loss_ids']:
                    encoder_inputs[key].append([piece[key]]*encoding_length)

        # decoder input ids
        decoder_inputs = {'decoder_input_ids': decoder_input_ids, 'loss_ids':loss_ids}
        decoder_inputs = self.truncate_decoder_inputs(decoder_inputs)

        encoder_inputs = self.truncate(encoder_inputs=encoder_inputs, num_special_times=1)

        # delete shortenable ids
        encoder_inputs.pop("shortenable_ids")
        encoder_inputs = self.concate_parts(input_dict=encoder_inputs)
        encoder_inputs = self.add_special_tokens(encoder_inputs=encoder_inputs)

        # create special input ids
        encoder_inputs['attention_mask'] = [1] *len(encoder_inputs['input_ids'])
        # padding
        encoder_inputs = self.padding(input_dict=encoder_inputs, max_len=self.max_seq_length, pad_id_for_inputs=self.tokenizer.pad_token_id)

        all_input_ids = {**encoder_inputs, **decoder_inputs}
        return all_input_ids

    def truncate_decoder_inputs(self, inputs):
        if self.decode_from_start:
            inputs['decoder_input_ids'].insert(0, self.tokenizer.sod_token_id)
            inputs['loss_ids'].insert(0, 0)

        for key in inputs:
            inputs[key] = inputs[key][:self.decoder_max_length - 1]

        if self.predict_eos:
            inputs['decoder_input_ids'].append(self.tokenizer.eos_token_id)
            inputs['loss_ids'].append(1)
        inputs = self.padding(inputs, max_len = self.decoder_max_length, pad_id_for_inputs=self.tokenizer.pad_token_id)
        return inputs










from pickle import FALSE
from torch.utils.data.sampler import RandomSampler
from transformers.configuration_utils import PretrainedConfig
from transformers import GenerationMixin
import torch
import torch.nn as nn
from torch.utils.data import Dataset
from typing import *
from SohuPrompt.data_utils import InputExample, InputFeatures
from torch.utils.data._utils.collate import default_collate
from tqdm.std import tqdm
from transformers.tokenization_utils import PreTrainedTokenizer
from transformers.utils.dummy_pt_objects import PreTrainedModel
from SohuPrompt.plms.utils import TokenizerWrapper
from SohuPrompt.prompt_base import Template, Verbalizer
from SohuPrompt.plms.glm import ChatGLMForConditionalGeneration
from collections import defaultdict
from SohuPrompt.utils import round_list, signature
import numpy as np
from torch.utils.data import DataLoader
from yacs.config import CfgNode
from SohuPrompt.utils.logging import logger
from transformers import AdamW, get_linear_schedule_with_warmup


class PromptDataLoader(object):
    r"""
    PromptDataLoader wraps the original dataset. The input data is firstly wrapped with the
    prompt's template, and then is tokenized by a wrapperd-tokenizer.
    Args:
        dataset (:obj:`Dataset` or :obj:`List`): Either a DatasetObject or a list containing the input examples.
        template (:obj:`Template`): A derived class of :obj:`Template`
        tokenizer (:obj:`PretrainedTokenizer`): The pretrained tokenizer.
        tokenizer_wrapper_class (:cls:`TokenizerWrapper`): The class of tokenizer wrapper.
        max_seq_length (:obj:`int`, optional): The max sequence length of the input ids. It's used to truncate sentences.
        batch_size (:obj:`int`, optional): The batch_size of data loader
        teacher_forcing (:obj:`bool`, optional): Whether to fill the mask with target text. Set to true in training generation model.
        decoder_max_length (:obj:`int`, optional): the decoder maximum length of an encoder-decoder model.
        predict_eos_token (:obj:`bool`, optional): Whether to predict the <eos> token. Suggest to set to true in generation.
        truncate_method (:obj:`bool`, optional): the truncate method to use. select from `head`, `tail`, `balanced`.
        kwargs  :Other kwargs that might be passed into a tokenizer wrapper.
    """

    def __init__(self,
                 dataset: Union[Dataset, List],
                 template: Template,
                 tokenizer_wrapper: Optional[TokenizerWrapper] = None,
                 tokenizer: PreTrainedTokenizer = None,
                 tokenizer_wrapper_class=None,
                 verbalizer: Optional[Verbalizer] = None,
                 max_seq_length: Optional[str] = 512,
                 batch_size: Optional[int] = 1,
                 shuffle: Optional[bool] = False,
                 teacher_forcing: Optional[bool] = False,
                 decoder_max_length: Optional[int] = -1,
                 predict_eos_token: Optional[bool] = False,
                 truncate_method: Optional[str] = "tail",
                 drop_last: Optional[bool] = False,
                 **kwargs,
                 ):

        assert hasattr(dataset, "__iter__"), f"The dataset must have __iter__ method. dataset is {dataset}"
        assert hasattr(dataset, "__len__"), f"The dataset must have __len__ method. dataset is {dataset}"

        self.raw_dataset = dataset
        self.wrapped_dataset = []
        self.tensor_dataset = []
        self.template = template
        self.verbalizer = verbalizer
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.teacher_forcing = teacher_forcing

        if tokenizer_wrapper is None:
            if tokenizer_wrapper_class is None:
                raise RuntimeError("Either wrapped_tokenizer or tokenizer_wrapper_class should be specified.")
            if tokenizer is None:
                raise RuntimeError("No tokenizer specified to instantiate tokenizer_wrapper.")

            tokenizer_wrapper_init_keys = signature(tokenizer_wrapper_class.__init__).args
            prepare_kwargs = {
                "max_seq_length": max_seq_length,
                "truncate_method": truncate_method,
                "decoder_max_length": decoder_max_length,
                "predict_eos_token": predict_eos_token,
                "tokenizer": tokenizer,
                **kwargs,
            }

            to_pass_kwargs = {key: prepare_kwargs[key] for key in prepare_kwargs if key in tokenizer_wrapper_init_keys}
            self.tokenizer_wrapper = tokenizer_wrapper_class(**to_pass_kwargs)
        else:
            self.tokenizer_wrapper = tokenizer_wrapper

        # check the satisfiability of each component
        assert hasattr(self.template, 'wrap_one_example'), "Your prompt has no function variable named wrap_one_example"

        # process
        self.wrap()
        self.tokenize()

        if self.shuffle:
            sampler = RandomSampler(self.tensor_dataset)
        else:
            sampler = None

        self.dataloader = DataLoader(self.tensor_dataset, batch_size=self.batch_size, sampler=sampler, collate_fn=InputFeatures.collate_fct, drop_last=drop_last, )

    def wrap(self):
        r"""A simple interface to pass the examples to prompt, and wrap the text with template.
        """
        if isinstance(self.raw_dataset, Dataset) or isinstance(self.raw_dataset, List):
            assert len(self.raw_dataset) > 0, 'The dataset to be wrapped is empty.'
            # for idx, example in tqdm(enumerate(self.raw_dataset),desc='Wrapping'):
            for idx, example in enumerate(self.raw_dataset):
                if self.verbalizer is not None and hasattr(self.verbalizer, 'wrap_one_example'):  # some verbalizer may also process the example.
                    example = self.verbalizer.wrap_one_example(example)
                wrapped_example = self.template.wrap_one_example(example)
                self.wrapped_dataset.append(wrapped_example)
        else:
            raise NotImplementedError

    def tokenize(self) -> None:
        r"""Pass the wrapped text into a prompt-specialized tokenizer,
           the true PretrainedTokenizer inside the tokenizer is flexible, e.g. AlBert, Bert, T5,...
        """
        for idx, wrapped_example in tqdm(enumerate(self.wrapped_dataset), desc='tokenizing'):
            # for idx, wrapped_example in enumerate(self.wrapped_dataset):
            inputfeatures = InputFeatures(**self.tokenizer_wrapper.tokenize_one_example(wrapped_example, self.teacher_forcing), **wrapped_example[1]).to_tensor()
            self.tensor_dataset.append(inputfeatures)

    def __len__(self):
        return len(self.dataloader)

    def __iter__(self, ):
        return self.dataloader.__iter__()


class PromptModel(nn.Module):
    r'''``PromptModel`` is the encapsulation of ``Template`` and the ``pre-trained model``,
    with OpenPrompt, these modules could be flexibly combined. And this class is the base class of ``PromptForClassification`` and ``PromptForGeneration``
    Args:
        plm (:obj:`PreTrainedModel`): The pre-trained language model for the current prompt-learning task.
        template (:obj:`Template`): The ``Template`` object to warp the input data.
        freeze_plm (:obj:`bool`): whether or not to freeze the pretrained language model
        plm_eval_mode (:obj:`bool`): this is a stronger freezing mode than freeze_plm, i.e. the dropout of the model is turned off. No matter whether the other part is set to train.
    '''

    def __init__(self,
                 plm: PreTrainedModel,
                 template: Template,
                 freeze_plm: bool = False,
                 plm_eval_mode: bool = False,
                 ):
        super().__init__()
        self.plm = plm
        self.template = template
        self.freeze_plm = freeze_plm
        self.plm_eval_mode = plm_eval_mode
        if freeze_plm:
            for param in self.plm.parameters():
                param.requires_grad = False
        if plm_eval_mode:
            self.plm.eval()
            for param in self.plm.parameters():
                param.requires_grad = False

        # get model's forward function's keywords
        self.forward_keys = signature(self.plm.forward).args

        self._prepare_main_input_name()

    def _prepare_main_input_name(self):
        model = self.plm
        if hasattr(model, "encoder") and hasattr(model.encoder, "main_input_name"):
            if model.encoder.main_input_name != model.main_input_name:
                main_input_name = model.encoder.main_input_name
            else:
                main_input_name = model.main_input_name
        else:
            main_input_name = getattr(model, "main_input_name", "input_ids")
        self.main_input_name = main_input_name

    def train(self, mode: bool = True):
        if not isinstance(mode, bool):
            raise ValueError("training mode is expected to be boolean")
        self.training = mode
        for name, module in self.named_children():
            if not (self.plm_eval_mode and 'plm' in name and mode):
                module.train(mode)
        return self

    def forward(self, batch: Union[Dict, InputFeatures]) -> torch.Tensor:
        r"""
        This is a forward method to make wrapped input data go through the model, and return the output logits.
        Typically, this function aims to predict the ``<mask>`` position.
        Args:
            batch (:obj:`Union[Dict, InputFeatures]`): The input features of batchified data sequences.
        """
        batch = self.template.process_batch(batch)
        input_batch = {key: batch[key] for key in batch if key in self.forward_keys}
        outputs = self.plm(**input_batch, output_hidden_states=True)
        outputs = self.template.post_processing_outputs(outputs)
        return outputs

    def prepare_model_inputs(self, batch: Union[Dict, InputFeatures]) -> Dict:
        r"""Will be used in generation
        """
        batch = self.template.process_batch(batch)
        input_batch = {key: batch[key] for key in batch if key in self.forward_keys}
        return input_batch


class PromptForClassification(nn.Module):
    r'''``PromptModel`` with a classification head on top. The classification head will map
    the logits in all position of the sequence (return value of a ``PromptModel``) into the
    logits of the labels, using a verbalizer.
    Args:
        plm (:obj:`PretrainedModel`): A pre-traiend model you decide to use for classification, e.g. BERT.
        template (:obj:`Template`): A ``Template`` object you use to wrap the input text for classification, e.g. ``ManualTemplate``.
        verbalizer (:obj:`Verbalizer`): A ``Verbalizer`` object you use to project the labels to label words for classification, e.g. ``ManualVerbalizer``.
        freeze_plm (:obj:`bool`): whether or not to freeze the pretrained language model
        plm_eval_mode (:obj:`bool`): this is a stronger freezing mode than freeze_plm, i.e. the dropout of the model is turned off. No matter whether the other part is set to train.
    '''

    def __init__(self,
                 plm: PreTrainedModel,
                 template: Template,
                 verbalizer: Verbalizer,
                 freeze_plm: bool = False,
                 plm_eval_mode: bool = False
                 ):
        super().__init__()
        self.prompt_model = PromptModel(plm, template, freeze_plm, plm_eval_mode)
        self.verbalizer = verbalizer

    @property
    def plm(self):
        return self.prompt_model.plm

    @property
    def template(self):
        return self.prompt_model.template

    @property
    def device(self, ):
        r"""Register the device parameter."""
        return self.plm.device

    def extract_at_mask(self,
                        outputs: torch.Tensor,
                        batch: Union[Dict, InputFeatures]):
        r"""Get outputs at all <mask> token
        E.g., project the logits of shape
        (``batch_size``, ``max_seq_length``, ``vocab_size``)
        into logits of shape (if num_mask_token > 1)
        (``batch_size``, ``num_mask_token``, ``vocab_size``)
        or into logits of shape (if ``num_mask_token`` = 1)
        (``batch_size``, ``vocab_size``).
        Args:
            outputs (:obj:`torch.Tensor`): The original outputs (maybe process by verbalizer's
                 `gather_outputs` before) etc. of the whole sequence.
            batch (:obj:`Union[Dict, InputFeatures]`): The original batch
        Returns:
            :obj:`torch.Tensor`: The extracted outputs of ``<mask>`` tokens.
        """
        outputs = outputs[torch.where(batch['loss_ids'] > 0)]
        outputs = outputs.view(batch['loss_ids'].shape[0], -1, outputs.shape[1])
        if outputs.shape[1] == 1:
            outputs = outputs.view(outputs.shape[0], outputs.shape[2])
        return outputs

    def forward(self, batch: Union[Dict, InputFeatures]) -> torch.Tensor:
        r"""
        Get the logits of label words.
        Args:
            batch (:obj:`Union[Dict, InputFeatures]`): The original batch
        Returns:
            :obj:`torch.Tensor`: The logits of the label words (obtained by the current verbalizer).
        """
        outputs = self.prompt_model(batch)
        outputs = self.verbalizer.gather_outputs(outputs)
        if isinstance(outputs, tuple):
            outputs_at_mask = [self.extract_at_mask(output, batch) for output in outputs]
        else:
            outputs_at_mask = self.extract_at_mask(outputs, batch)
        label_words_logits = self.verbalizer.process_outputs(outputs_at_mask, batch=batch)
        return label_words_logits

    def predict(self):
        pass

    def forward_without_verbalize(self, batch: Union[Dict, InputFeatures]) -> torch.Tensor:
        outputs = self.prompt_model(batch)
        outputs = self.verbalizer.gather_outputs(outputs)
        outputs_at_mask = self.extract_at_mask(outputs, batch)
        return outputs_at_mask

    @property
    def tokenizer(self):
        r'''Utility property, to get the tokenizer more easily.
        '''
        return self.verbalizer.tokenizer

    ##  We comment this code since it conflict with [OpenDelta](https://github.com/thunlp/OpenDelta)
    # def state_dict(self, *args, **kwargs):
    #     """ Save the model using template, plm and verbalizer's save methods."""
    #     _state_dict = {}
    #     if not self.prompt_model.freeze_plm:
    #         _state_dict['plm'] = self.plm.state_dict(*args, **kwargs)
    #     _state_dict['template'] = self.template.state_dict(*args, **kwargs)
    #     _state_dict['verbalizer'] = self.verbalizer.state_dict(*args, **kwargs)
    #     return _state_dict

    # def load_state_dict(self, state_dict, *args, **kwargs):
    #     """ Load the model using template, plm and verbalizer's load methods."""
    #     if 'plm' in state_dict and not self.prompt_model.freeze_plm:
    #         self.plm.load_state_dict(state_dict['plm'], *args, **kwargs)
    #     self.template.load_state_dict(state_dict['template'], *args, **kwargs)
    #     self.verbalizer.load_state_dict(state_dict['verbalizer'], *args, **kwargs)

    def parallelize(self, device_map=None):
        r"""Parallelize the model across device
        """
        if hasattr(self.plm, "parallelize"):
            self.plm.parallelize(device_map)
            self.device_map = self.plm.device_map
            self.template.cuda()
            self.verbalizer.cuda()
        else:
            raise NotImplementedError("parallelize method was not implemented for this plm.")

    def deparallelize(self):
        r"""Deparallelize the model across device
        """
        if hasattr(self.plm, "deparallelize"):
            self.plm.deparallelize()
            self.device_map = None
            self.template.cpu()
            self.verbalizer.cpu()
        else:
            raise NotImplementedError("parallelize method was not implemented for this plm.")


class PromptForGeneration(nn.Module, GenerationMixin):
    r'''``PromptModel`` with generation loss calculation and generation utils integrated.
    Args:
        plm (:obj:`PretrainedModel`): A pre-traiend model you decide to use for generation, e.g. GPT.
        template (:obj:`Template`): A ``Template`` object you use to wrap the input text for classification, e.g. ``PrefixTuningTemplate``.
        tokenizer (:obj:`Tokenizer`): A ``Tokenizer`` of the current model.
        gen_config (:obj:`CfgNode`): The generation configs to pass into `GenerationMixin.generate <https://huggingface.co/transformers/_modules/transformers/generation_utils.html#GenerationMixin.generate>`_
        freeze_plm (:obj:`bool`): whether or not to freeze the pretrained language model
        plm_eval_mode (:obj:`bool`): this is a stronger freezing mode than freeze_plm, i.e. the dropout of the model is turned off. No matter whether the other part is set to train.
    '''

    def __init__(self,
                 plm: PreTrainedModel,
                 template: Template,
                 freeze_plm: bool = False,
                 plm_eval_mode: bool = False,
                 gen_config: Optional[CfgNode] = None,
                 tokenizer: Optional[PreTrainedTokenizer] = None,
                 ):
        super().__init__()
        self.freeze_plm = freeze_plm
        if tokenizer is None:
            assert template.tokenizer is not None, "Tokenizer can't be set from input args or template"
            self.tokenizer = template.tokenizer
        else:
            self.tokenizer = tokenizer
        self.prompt_model = PromptModel(plm, template, freeze_plm, plm_eval_mode)

        self.loss_fct = nn.CrossEntropyLoss(reduction='none')
        self.config = plm.config
        if gen_config:
            for key in gen_config:
                setattr(self.config, key, gen_config[key])
        self.in_generation_function = False

        self.main_input_name = self.prompt_model.main_input_name  # for transformers 4.17.0 and higher.

    @property
    def generation_config(self):
        return self.plm.generation_config

    @property
    def plm(self):
        return self.prompt_model.plm

    @property
    def template(self):
        return self.prompt_model.template

    @property
    def device(self):
        return self.plm.device

    def can_generate(self):
        if hasattr(self.prompt_model.plm, "can_generate"):
            return self.prompt_model.plm.can_generate
        else:
            raise AttributeError("the corresponding plm does not have `can_generate` attribute")

    def shift_logits_and_labels(self, logits, loss_ids, reference_ids):
        r"""
        Left shift the label, and make label of the positions that are
        not loss position to -100, which is the ignore index in pytorch's
        loss function.
        Args:
            logits (:obj:`torch.Tensor`):
            batch (:obj:`InputFeatures`): The input features of batchified data sequences.
        Returns:
            shift_logits (:obj:`torch.Tensor`):
            shift_input_ids (:obj:`List[int]`):
        """
        shift_logits = logits[..., :-1, :].contiguous()
        shift_loss_ids = loss_ids[..., 1:].contiguous()
        shift_input_ids = reference_ids[..., 1:].contiguous()
        shift_input_ids = torch.where(shift_loss_ids > 0, shift_input_ids, -100)
        return shift_logits, shift_input_ids

    def forward(self, *args, **kwargs):
        r"""In generation process, it will use the plm's forward function.
        This is because, in the first step we will directly call the process_batch function to
        generate initial input with the template, after that the all template
        have been processed into the past_key_value,
        then we can use the normal generation function.
        In learning process, the forward is linked to ``_forward`` functions.
        in which the loss will be calculated for all the positions in the same time.
        """
        if self.in_generation_function:
            return self.plm.forward(*args, **kwargs)
        else:
            return self._forward(*args, **kwargs)

    def _forward(self, batch: Union[Dict, InputFeatures], label_smoother, is_chat_glm) -> torch.Tensor:
        r"""
        This is the forward method of the training of generation in prompt-learning framework.
        Args:
            batch (:obj:`Union[Dict, InputFeatures]`): The input features of batchified data sequences.
        Returns:
            loss(:obj:torch.Tensor): The loss of the current generation procedure.
        """
        if self.config.is_encoder_decoder:
            reference_ids = batch['decoder_input_ids']
        else:
            reference_ids = batch['input_ids']  # in case in some template, these field is dropped
        if is_chat_glm:
            model_inputs = batch
            model_inputs.update(self.prompt_model.plm.prepare_inputs_for_generation(input_ids=batch['input_ids']))
        else:
            model_inputs = batch
        outputs = self.prompt_model(model_inputs)
        logits = outputs.logits
        logits, labels = self.shift_logits_and_labels(logits, batch['loss_ids'], reference_ids)
        if label_smoother is not None:
            pass
        else:
            batch_size, seq_len, vocab_size = logits.shape
            loss = self.loss_fct(logits.view(-1, logits.size(-1)), labels.view(-1))
        loss = loss.view(batch_size, -1).sum(dim=-1)  # TODO support more objectives
        loss = loss.mean()
        return loss

    def _validate_model_kwargs(self, model_kwargs: Dict[str, Any]):
        """skip model kwargs validation, as all arguments will be filtered before feeding into plm in PromptModel"""
        return

    def generate(self, batch: Union[Dict, InputFeatures], verbose: Optional[bool] = False, **generation_kwargs):
        r""" This function wraps the generate() methods in parent class ``GenerationMixin``.
        Forward uses the ``PretrainedModel``'s forward method.
        generation_kwargs include all the parameters that are passed in to
        ``transformers.generation_util.GenerationMixin.generate``
        Args:
            batch (:obj:`Union[Dict, InputFeatures]`): The input features of batchified data sequences.
            verbose (:obj:`Optional[bool]`): Set to true to verbose the generated sentence.
        Returns:
            output_sequences (:obj:`List[torch.Tensor]`): The raw sequences generated by the generation model.
            generated_sentences (:obj:`List[torch.Tensor]`): The generated sentences that have been post-processed.
        """
        input_generation_kwargs = generation_kwargs
        if self.config.is_encoder_decoder:
            loss_ids_start = batch['loss_ids'].argmax(dim=-1)
            assert loss_ids_start.min() == loss_ids_start.max(), "The generation start from different position in a batch."
            batch['decoder_input_ids'] = batch['decoder_input_ids'][:, :loss_ids_start.min() + 1]
            input_length = batch['decoder_input_ids'].size(1)
            batch_size = batch['decoder_input_ids'].size(0)

            self.generate_ith_token = 0
            self.in_generation_function = True
            output_sequences = super().generate(**batch, **input_generation_kwargs, pad_token_id=self.tokenizer.pad_token_id, eos_token_id=self.tokenizer.eos_token_id)
            self.in_generation_function = False
            output_sequences = output_sequences.cpu().tolist()
            generated_sentences = self.post_processing(output_sequences=output_sequences, input_lengths=input_length)
        else:
            input_length = batch['input_ids'].size(1)
            batch_size = batch['input_ids'].size(0)

            # Currently huggingface transformers only support single sample generation, or padding to the left (instead of the right).
            # because it will only extract the last position of the output
            # generate one_by_one
            if 'input_ids_len' in batch:
                input_real_lens = batch['input_ids_len']
            else:
                input_real_lens = torch.sum((batch['input_ids'] != self.tokenizer.pad_token_id).to(torch.int), dim=-1)
            output_sequences = []
            for instance_id in range(batch_size):
                # remove the pad token
                instance = {key: batch[key][instance_id:instance_id + 1][:, :input_real_lens[instance_id]] for key in batch if isinstance(batch[key], torch.Tensor) and batch[key].shape[:2] == torch.Size([batch_size, input_length])}
                self.generate_ith_token = 0
                self.in_generation_function = True
                output_sequence = super().generate(**instance, **input_generation_kwargs, pad_token_id=self.tokenizer.pad_token_id, eos_token_id=self.tokenizer.eos_token_id)
                self.in_generation_function = False
                output_sequences.extend(output_sequence.cpu().tolist())  # TODO: to support generate multiple sentence
            generated_sentences = self.post_processing(output_sequences=output_sequences, input_lengths=input_real_lens.cpu().tolist())
        if verbose:
            logger.info(f"Generated:{generated_sentences}")
        return output_sequences, generated_sentences

    def post_processing(self, output_sequences, input_lengths):
        r"""
            Post-process the sequences generated by the generation model.
            Args:
                output_sequences (:obj:`torch.Tensor`): The raw sequences generated by the generation model.
                input_lengths (:obj:`int` or `list`): The length(s) of the input sequence.
            Returns:
                :obj:`List`: The generated sentences that have been post-processed.
        """
        generated_sentences = []
        if type(input_lengths) == int:
            input_lengths = [input_lengths] * len(output_sequences)
        for sent_id, seq in enumerate(output_sequences):
            seq = seq[input_lengths[sent_id]:]

            if hasattr(self.tokenizer, "eos_token") and self.tokenizer.eos_token is not None:
                text_output = self.tokenizer.decode(seq, clean_up_tokenization_spaces=True, skip_special_tokens=False)
                idx = text_output.find(self.tokenizer.eos_token)
                if idx >= 0:
                    text_output = text_output[:idx]
            else:
                text_output = self.tokenizer.decode(seq, clean_up_tokenization_spaces=True, skip_special_tokens=True)
            text_output = text_output.strip()
            generated_sentences.append(text_output)
        return generated_sentences

    def prepare_inputs_for_generation(self, input_ids: Optional[torch.Tensor] = None, **model_kwargs):
        r"""This function wraps the ``prepare_inputs_for_generation`` function in the huggingface transformers.
        When the `past` not in model_kwargs, we prepare the input from scratch.
        When `past` is in model_kwargs, we don't need to prepare the template wrapped input,
        instead we use the inner pretrain_models' function to prepare the next step's input.
        `model_kwargs` includes all the argument passed in the `batch`: InputFeatures, except ``input_ids``
        , as long as they do not conflict with keywords in ``generation_kwargs``.    if 'past' not in model_kwargs: # the past_key_value not in model_kwargs, then we need to prepare input from scrath
        , as long as they do not conflict with keywords in ``generation_kwargs``.
        Args:
            input_ids(:obj:`torch.Tensor`): Indices of input sequence tokens in the vocabulary.
        """
        if self.generate_ith_token == 0 and 'encoder_outputs' not in model_kwargs and not isinstance(self.plm, ChatGLMForConditionalGeneration):  # generating the first token in decoder only setting.
            batch = InputFeatures(input_ids=input_ids, **model_kwargs)
            model_inputs = self.prompt_model.prepare_model_inputs(batch)
            # check the compatibility for more models. Having checked gpt2, T5
        elif isinstance(self.plm, ChatGLMForConditionalGeneration):
            model_kwargs['input_ids'] = input_ids
            model_inputs = model_kwargs
            model_inputs.pop('attention_mask')
            # input_ids生成position_ids和attention_mask
            model_inputs.update(self.plm.prepare_inputs_for_generation(**model_inputs))
            if model_inputs['past_key_values'] is None:
                # input_ids生成input_embs
                model_inputs.update(self.prompt_model.prepare_model_inputs(model_inputs))
            model_inputs = {key: model_inputs[key] for key in model_inputs if key in self.prompt_model.forward_keys}
        else:  # generating the subsequence generation can use the default setting
            model_inputs = self.plm.prepare_inputs_for_generation(input_ids, **model_kwargs)
            
        self.last_model_inputs = model_inputs  # to update the model_kwargs in _update_model_kwargs_for_generation, in-place operation.
        return model_inputs

    def _update_model_kwargs_for_generation(self, outputs, model_kwargs: Dict[str, Any], is_encoder_decoder: bool = False) -> Dict[str, Any]:
        r""" The parents class's ``_update_model_kwargs_for_generation`` method will
        add ``past_key_values`` to model_kwargs, and update ``token_type_ids``, and ``attention_mask_ids``.
        In case some of the model_kwargs are modified in the prepare_inputs_for_generation function
        and should be used as the subsequent model_kwargs, we update these kwargs before the parent class
        call.
        Other updates should be added here after the parent's function call.
        Args:
            outputs (:obj:`torch.Tensor`):
            is_encoder_decoder (:obj:`bool`, defaults to False):
        """
        if isinstance(self.prompt_model.plm, ChatGLMForConditionalGeneration):
            model_kwargs = self.prompt_model.plm._update_model_kwargs_for_generation(outputs=outputs, model_kwargs=model_kwargs, is_encoder_decoder=is_encoder_decoder)
        else:
            if self.generate_ith_token == 0:
                for key in self.last_model_inputs:
                    if key in model_kwargs:
                        model_kwargs[key] = self.last_model_inputs[key]
            model_kwargs = super()._update_model_kwargs_for_generation(outputs=outputs, model_kwargs=model_kwargs, is_encoder_decoder=is_encoder_decoder)
            self.generate_ith_token += 1
        return model_kwargs

    def _prepare_encoder_decoder_kwargs_for_generation(
            self, input_ids: torch.LongTensor, model_kwargs, model_input_name: Optional[str] = None,
    ) -> Dict[str, Any]:
        r""" This function resemble the function in GeneraionMix
        Args:
            input_ids (:obj:`torch.LongTensor`) The input ids for
        """
        if "encoder_outputs" not in model_kwargs:
            # retrieve encoder hidden states
            encoder = self.plm.get_encoder()
            encoder_kwargs = {
                argument: value
                for argument, value in model_kwargs.items()
                if not (argument.startswith("decoder_") or argument.startswith("cross_attn") or argument.startswith("use_cache"))
            }
            model_input_name = model_input_name if model_input_name is not None else self.main_input_name
            batch = {model_input_name: input_ids, **encoder_kwargs}
            model_inputs = self.prompt_model.prepare_model_inputs(batch)  # This line differs from the orinigal code base, we should process the input
            # with our template, then pass it into the model.
            # some of the arguments may have been changed by the template,
            # e.g. the attention mask. Here we update the model_kwargs
            for key in model_kwargs:
                if key in model_inputs:
                    model_kwargs[key] = model_inputs[key]
            model_kwargs["encoder_outputs"] = encoder(return_dict=True, **model_inputs)
        return model_kwargs

    ##  We comment this code since it conflict with [OpenDelta](https://github.com/thunlp/OpenDelta)
    # def state_dict(self, *args, **kwargs):
    #     """ Save the model using template and plm's save methods. """
    #     _state_dict = {}
    #     if not self.prompt_model.freeze_plm:
    #         _state_dict['plm'] = self.plm.state_dict(*args, **kwargs)
    #     _state_dict['template'] = self.template.state_dict(*args, **kwargs)
    #     return _state_dict

    # def load_state_dict(self, state_dict, *args, **kwargs):
    #     """ Load the model using template and plm's load methods. """
    #     if 'plm' in state_dict and not self.prompt_model.freeze_plm:
    #         self.plm.load_state_dict(state_dict['plm'], *args, **kwargs)
    #     self.template.load_state_dict(state_dict['template'], *args, **kwargs)

    def _reorder_cache(self, past, beam_idx):
        r"""Use the plm's default _reorder_cache function
        """
        return self.plm._reorder_cache(past, beam_idx)

    def parallelize(self, device_map=None):
        r"""Parallelize the model across device
        """
        if hasattr(self.plm, "parallelize"):
            self.plm.parallelize(device_map)
            self.device_map = self.plm.device_map
        else:
            raise NotImplementedError("parallelize method was not implemented for this plm.")

    def deparallelize(self):
        r"""Deparallelize the model across device
        """
        if hasattr(self.plm, "deparallelize"):
            self.plm.deparallelize()
            self.device_map = None
        else:
            raise NotImplementedError("parallelize method was not implemented for this plm.")




import copy
import json
import pickle
from typing import *

import torch
from torch.utils.data._utils.collate import default_collate
from SohuPrompt.utils.logging import logger

from typing import Union


class InputExample(object):
    """A raw input example consisting of segments of text,
    a label for classification task or a target sequence of generation task.
    Other desired information can be passed via meta.

    Args:
        guid (:obj:`str`, optional): A unique identifier of the example.
        text_a (:obj:`str`, optional): The placeholder for sequence of text.
        text_b (:obj:`str`, optional): A secend sequence of text, which is not always necessary.
        label (:obj:`int`, optional): The label id of the example in classification task.
        tgt_text (:obj:`Union[str,List[str]]`, optional):  The target sequence of the example in a generation task..
        meta (:obj:`Dict`, optional): An optional dictionary to store arbitrary extra information for the example.
    """

    def __init__(self,
                 guid = None,
                 text_a = "",
                 text_b = "",
                 text_c = "",
                 label = None,
                 meta: Optional[Dict] = None,
                 tgt_text: Optional[Union[str,List[str]]] = None
                ):

        self.guid = guid
        self.text_a = text_a
        self.text_b = text_b
        self.text_c = text_c
        self.label = label
        self.meta = meta if meta else {}
        self.tgt_text = tgt_text

    def __repr__(self):
        return str(self.to_json_string())

    def to_dict(self):
        r"""Serialize this instance to a Python dictionary."""
        output = copy.deepcopy(self.__dict__)
        return output

    def to_json_string(self):
        r"""Serialize this instance to a JSON string."""
        return json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"

    def keys(self, keep_none=False):
        return [key for key in self.__dict__.keys() if getattr(self, key) is not None]

    @staticmethod
    def load_examples(path: str) -> List['InputExample']:
        """Load a set of input examples from a file"""
        with open(path, 'rb') as fh:
            return pickle.load(fh)

    @staticmethod
    def save_examples(examples: List['InputExample'], path: str) -> None:
        """Save a set of input examples to a file"""
        with open(path, 'wb') as fh:
            pickle.dump(examples, fh)


class InputFeatures(dict):
    """
    The class for input to the PLM and Prompts. To make users explicitly know the available keys,
    we define a dict with a set of predefined possible keys. The default value to any key is None.
    When use it as a dict, all the keys whose values are None are invisible.

    This class support most of the dict's operation (See Examples). It can also be consumed by
    pytorch's default_collate in DataLoader.
    Also a :py:meth:`to_tensor()` method is build to convert the values into torch.Tensor for torch's input.

    Examples:

    ..  code-block:: python

        in_feat = InputFeatures(**{'input_ids':[1,4,5], 'soft_token_ids': [3,4,5]})  # init from dict
        print(in_feat.keys())       # ['input_ids, 'soft_token_ids']
        in_feat['label'] = 3        # can assign value like normal dict
        print(in_feat.keys())       # ['input_ids','label', 'soft_token_ids'] (Note that it's also ordered)
        print(in_feat['label'])     # 3
        in_feat['alice'] = 0        # KeyError: Key alice not in predefined set of keys
        in_feat.values()            # [[1,4,5], 3, [3,4,5]]  (Note that it's also ordered)
        [in_feat[key] for key in in_feat]   # [[1,4,5], 3, [3,4,5]]
        new_dict= {**in_feat, 'new_key':2}  # new_dict is {'input_ids': [1, 4, 5], 'label': 3, 'soft_token_ids': [3, 4, 5], 'new_key': 2}

    Args:
        input_ids: Indices of input sequence tokens in the vocabulary.
        attention_mask: Mask to avoid performing attention on padding token indices.
            Mask values selected in ``[0, 1]``: Usually ``1`` for tokens that are NOT MASKED, ``0`` for MASKED (padded)
            tokens.
        token_type_ids: (Optional) Segment token indices to indicate first and second
            portions of the inputs. Only some models use them.
        label: (Optional) Label corresponding to the input. Int for classification problems,
            float for regression problems.
    """
    tensorable_keys = ['input_ids', 'inputs_embeds', 'attention_mask', 'token_type_ids', 'label',
        'decoder_input_ids', 'decoder_inputs_embeds', 'soft_token_ids',
        'past_key_values', 'loss_ids']
    all_keys = ['input_ids', 'inputs_embeds', 'attention_mask', 'token_type_ids', 'label',
        'decoder_input_ids', 'decoder_inputs_embeds', 'soft_token_ids',
        'past_key_values', 'loss_ids', 'guid', 'tgt_text', 'encoded_tgt_text', 'input_ids_len']
    non_tensorable_keys = []

    def __init__(self,
                input_ids: Optional[Union[List, torch.Tensor]] = None,
                inputs_embeds: Optional[torch.Tensor] = None,
                attention_mask: Optional[Union[List[int], torch.Tensor]] = None,
                token_type_ids: Optional[Union[List[int], torch.Tensor]] = None,
                label: Optional[Union[int, torch.Tensor]] = None,
                decoder_input_ids: Optional[Union[List, torch.Tensor]] = None,
                decoder_inputs_embeds: Optional[torch.Tensor] = None,
                soft_token_ids: Optional[Union[List, torch.Tensor]] = None,
                past_key_values: Optional[torch.Tensor] = None,  # for prefix_tuning
                loss_ids: Optional[Union[List, torch.Tensor]] = None,
                guid: Optional[str] = None,
                tgt_text: Optional[str] = None,
                use_cache: Optional[bool] = None,
                encoded_tgt_text: Optional[str] = None,
                input_ids_len: Optional[int] = None,
                **kwargs):

        self.input_ids = input_ids
        self.inputs_embeds = inputs_embeds
        self.attention_mask = attention_mask
        self.token_type_ids = token_type_ids
        self.label = label
        self.decoder_input_ids = decoder_input_ids
        self.decoder_inputs_embeds = decoder_inputs_embeds
        self.soft_token_ids = soft_token_ids
        self.past_key_values = past_key_values
        self.loss_ids = loss_ids
        self.guid = guid
        self.tgt_text = tgt_text
        self.encoded_tgt_text = encoded_tgt_text
        self.use_cache = use_cache
        self.input_ids_len = input_ids_len

        for k in kwargs.keys():
            logger.warning("Your are passing an unexpected key words: {} to InputFeatures, might yield unexpected behaviours!".format(k))
            setattr(self, k, kwargs[k])

    @classmethod
    def add_tensorable_keys(cls, *args):
        cls.tensorable_keys.extend(args)

    @classmethod
    def add_not_tensorable_keys(cls, *args):
        cls.not_tensorable_keys.extend(args)

    @classmethod
    def add_keys(cls, *args):
        cls.all_keys.extend(args)

    def __repr__(self):
        return str(self.to_json_string())

    def __len__(self):
        return len(self.keys())

    def to_tensor(self, device: str = 'cuda'):
        """inplace operation, convert all tensorable features into :obj:`torch.tensor`"""
        for key in self.tensorable_keys:
            value = getattr(self, key)
            if value is not None:
                #print(key,value)
                setattr(self, key, torch.tensor(value))
        return self

    def to(self, device: str = "cuda:0"):
        r"""move the tensor keys to runtime device, such as gpu:0
        """
        for key in self.tensorable_keys:
            value = getattr(self, key)
            if value is not None:
                setattr(self, key, value.to(device))
        return self

    def cuda(self, device: str = "cuda:0"):
        r"""mimic the tensor behavior
        """
        return self.to(device)

    def to_json_string(self, keep_none=False):
        """Serializes this instance to a JSON string."""
        data = {}
        for key in self.all_keys:
            value = getattr(self, key)
            if isinstance(value, torch.Tensor):
                data[key] =  value.detach().cpu().tolist()
            elif value is None and keep_none:
                data[key] = None
            else:
                data[key] = value
        return json.dumps(data) + "\n"

    def keys(self, keep_none=False) -> List[str]:
        """get all keys of the InputFeatures

        Args:
            keep_none (:obj:`bool`, optional): whether to keep the predefined keys whose value is none. Defaults to False.

        Returns:
            :obj:`List[str]`: keys of the InputFeatures
        """
        if keep_none:
            return self.all_keys
        else:
            return [key for key in self.all_keys if getattr(self, key) is not None]

    def to_dict(self, keep_none=False) -> Dict[str, Any]:
        """get the dict of mapping from keys to values of the InputFeatures

        Args:
            keep_none (:obj:`bool`, optional): whether to keep the predefined keys whose value is none. Defaults to False.

        Returns:
            :obj:`Dict[str, Any]`: dict of mapping from keys to values of the InputFeatures
        """
        data = {}
        for key in self.all_keys:
            value = getattr(self, key)
            if value is not None:
                data[key] =  value
            elif value is None and keep_none:
                data[key] = None
        return data

    def __getitem__(self, key):
        return getattr(self, key)

    def __iter__(self):
        return iter(self.keys())

    def __setitem__(self, key, item):
        if key not in self.all_keys:
            raise KeyError("Key {} not in predefined set of keys".format(key))
        setattr(self, key, item)

    def values(self, keep_none=False) -> List[Any]:
        """get the values with respect to the keys  of the InputFeatures

        Args:
            keep_none (:obj:`bool`, optional): whether to keep the predefined keys whose value is none. Defaults to False.

        Returns:
            :obj:`List[Any]`: the values with respect to the keys of the InputFeatures
        """
        return [getattr(self, key) for key in self.keys(keep_none=keep_none)]

    def __contains__(self, key, keep_none=False):
        return key in self.keys(keep_none)

    def items(self,):
        """get the (key, value) pairs  of the InputFeatures

        Args:
            keep_none (:obj:`bool`, optional): whether to keep the predefined keys whose value is none. Defaults to False.

        Returns:
            :obj:`List[Any]`: the (key, value) pairs of the InputFeatures
        """
        return [(key, self.__getitem__(key)) for key in self.keys()]

    @staticmethod
    def collate_fct(batch: List):
        r'''
        This function is used to collate the input_features.

        Args:
            batch (:obj:`List[Union[Dict, InputFeatures]]`): A batch of the current data.

        Returns:
            :obj:`InputFeatures`: Return the :py:class:`~SohuPrompt.data_utils.data_utils.InputFeatures of the current batch of data.
        '''


        elem = batch[0]
        return_dict = {}
        for key in elem:
            if key == "encoded_tgt_text":
                return_dict[key] = [d[key] for d in batch]
            else:
                try:
                    return_dict[key] = default_collate([d[key] for d in batch])
                except:
                    print(f"key{key}\n d {[batch[i][key] for i in range(len(batch))]} ")

        return InputFeatures(**return_dict)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file contains the logic for loading data for all FewGLUE tasks.
"""


from SohuPrompt.data_utils.utils import InputExample

from SohuPrompt.data_utils.data_processor import DataProcessor
from datasets import load_dataset
from SohuPrompt.utils.logging import logger
from datasets import load_from_disk
import os

HUGGING_FACE_SCRIPTS = 'super_glue' # if you can not connect huggingface in your machine, you can download the scripts manually and change this line.
HUGGING_FACE_SCRIPTS = '../../huggingface_datasets/super_glue'
class SuperglueMultiRCProcessor(DataProcessor):
    def __init__(self):
        super().__init__()
        self.labels = ["No", "Yes"]

    def get_examples(self, data_dir, split):
        if split == "valid" or split == "dev":
            split = "validation"
        try:
            dataset = load_dataset(path=HUGGING_FACE_SCRIPTS, name='multirc', cache_dir=data_dir, split=split)
        except:
            dataset = load_from_disk(f"{data_dir}/super_glue.multirc")[split]
        return list(map(self.transform, dataset))

    def transform(self, example):
        text_a = example['paragraph']
        text_b = example['question']
        meta = {"answer": example["answer"]}
        label = int(example['label'])
        guid = "p{}-q{}-a{}".format(example['idx']['paragraph'], example['idx']['question'], example['idx']['answer'])
        return InputExample(guid = guid, text_a = text_a, text_b = text_b, meta = meta, label=label)

class SuperglueBoolQProcessor(DataProcessor):
    def __init__(self):
        super().__init__()
        self.labels = ["No", "Yes"]

    def get_examples(self, data_dir, split):
        if split == "valid" or split == "dev":
            split = "validation"
        try:
            dataset = load_dataset(path=HUGGING_FACE_SCRIPTS, name='boolq', cache_dir=data_dir, split=split)
        except:
            dataset = load_from_disk(f"{data_dir}/super_glue.boolq")[split]
        return list(map(self.transform, dataset))

    def transform(self, example):
        text_a = example['passage']
        text_b = example['question']
        label = int(example['label'])
        guid = "{}".format(example['idx'])
        return InputExample(guid = guid, text_a = text_a, text_b = text_b, label=label)

class SuperglueCBProcessor(DataProcessor):
    def __init__(self):
        super().__init__()
        self.labels = ["Entailment", "Not Entailment", "Neutral"]

    def get_examples(self, data_dir, split):
        if split == "valid" or split == "dev":
            split = "validation"
        try:
            dataset = load_dataset(path=HUGGING_FACE_SCRIPTS, name='cb', cache_dir=data_dir, split=split)
        except:
            dataset = load_from_disk(f"{data_dir}/super_glue.cb")[split]
        return list(map(self.transform, dataset))

    def transform(self, example):
        text_a = example['premise']
        text_b = example['hypothesis']
        label = int(example['label'])
        guid = "{}".format(example['idx'])
        return InputExample(guid = guid, text_a = text_a, text_b = text_b, label=label)


class SuperglueCOPAProcessor(DataProcessor):
    def __init__(self):
        super().__init__()
        self.labels = ["No", "Yes"]

    def get_examples(self, data_dir, split):
        if split == "valid" or split == "dev":
            split = "validation"
        try:
            dataset = load_dataset(path=HUGGING_FACE_SCRIPTS, name='copa', cache_dir=data_dir, split=split)
        except:
            dataset = load_from_disk(f"{data_dir}/super_glue.copa")[split]
        return list(map(self.transform, dataset))

    def transform(self, example):
        choice1 = example['choice1']
        choice2 = example['choice2']
        premise= example['premise']
        question = example['question']

        meta = {}
        meta['choice1'] = choice1
        meta['choice2'] = choice2
        meta['question'] = question
        label = int(example['label'])
        guid = "{}".format(example['idx'])
        return InputExample(guid = guid, text_a = premise, meta=meta, label=label)

class SuperglueRTEProcessor(DataProcessor):
    def __init__(self):
        super().__init__()
        self.labels = ["Entailment", "Not Entailment"]

    def get_examples(self, data_dir, split):
        if split == "valid" or split == "dev":
            split = "validation"
        try:
            dataset = load_dataset(path=HUGGING_FACE_SCRIPTS, name='rte', cache_dir=data_dir, split=split)
        except:
            dataset = load_from_disk(f"{data_dir}/super_glue.rte")[split]
        return list(map(self.transform, dataset))

    def transform(self, example):
        text_a = example['premise']
        text_b = example['hypothesis']
        label = int(example['label'])
        guid = "{}".format(example['idx'])
        return InputExample(guid = guid, text_a = text_a, text_b = text_b, label=label)

class SuperglueWiCProcessor(DataProcessor):
    def __init__(self):
        super().__init__()
        self.labels = ["Same", "Different"]

    def get_examples(self, data_dir, split):
        if split == "valid" or split == "dev":
            split = "validation"
        try:
            dataset = load_dataset(path=HUGGING_FACE_SCRIPTS, name='wic', cache_dir=data_dir, split=split)
        except:
            dataset = load_from_disk(f"{data_dir}/super_glue.wic")[split]
        return list(map(self.transform, dataset))

    def transform(self, example):
        meta = {}
        text_a = example["sentence1"]
        text_b = example["sentence2"]
        meta['word'] = example['word']
        label = int(example['label'])
        guid = "{}".format(example['idx'])
        return InputExample(guid = guid, text_a=text_a, text_b=text_b, meta=meta, label=label)

class SuperglueWSCProcessor(DataProcessor):
    def __init__(self):
        super().__init__()
        self.labels = ["Different", "Same"]

    def get_examples(self, data_dir, split):
        if split == "valid" or split == "dev":
            split = "validation"
        try:
            dataset = load_dataset(path=HUGGING_FACE_SCRIPTS, name='wsc', cache_dir=data_dir, split=split)
        except:
            dataset = load_from_disk(f"{data_dir}/super_glue.wsc")[split]
        return list(map(self.transform, dataset))

    def transform(self, example):
        modified_text =  example["text"].split()
        indices = sorted([example['span2_index'], example['span1_index']])
        for idx in indices[::-1]:
            modified_text.insert(idx+1, "*")
            modified_text.insert(idx, "*")
        modified_text = " ".join(modified_text)

        meta = {}
        meta['span1_text'] = example['span1_text']
        meta['span2_text'] = example['span2_text']
        label = int(example['label'])
        guid = "{}".format(example['idx'])
        return InputExample(guid = guid, text_a=modified_text, meta=meta, label=label)


class SuperglueRecordProcessor(DataProcessor):
    def __init__(self):
        super().__init__()
        self.labels = [None]

    def get_examples(self, data_dir, split):
        if split == "valid" or split == "dev":
            split = "validation"
        try:
            dataset = load_dataset(path=HUGGING_FACE_SCRIPTS, name='record', cache_dir=data_dir, split=split)
        except:
            dataset = load_from_disk(f"{data_dir}/super_glue.record")[split]
        return list(map(self.transform, dataset))

    def transform(self, example):
        meta = {}
        meta['passage'] = example["passage"].replace("\n", " ")
        meta['query'] = example["query"]
        meta['entities'] = ", ".join(example['entities'])
        if len(example['answers'])>0:
            meta['answers'] = example['answers'][0]
        else:
            meta['answers'] = ''
        guid = "{}".format(example['idx'])
        return InputExample(guid = guid, meta=meta, label=0)


class YahooAnswersTopicsProcessor(DataProcessor):
    def __init__(self):
        super().__init__()
        self.labels = [
                        "Society & Culture",
                        "Science & Mathematics",
                        "Health",
                        "Education & Reference",
                        "Computers & Internet",
                        "Sports",
                        "Business & Finance",
                        "Entertainment & Music",
                        "Family & Relationships",
                        "Politics & Government",
                    ]

    def get_examples(self, data_dir, split):
        if split == "valid" or split == "dev":
            split = "train"# "validation"
        try:
            dataset = load_dataset("yahoo_answers_topics",split=split) # If you have network issues, we use the manually downloaded datasets.
        except:
            dataset = load_from_disk(f"{data_dir}")#[split]
            dataset = dataset[split]
        return list(map(self.transform, dataset))

    def transform(self, example):
        meta = {}
        text_a = example["question_title"]
        text_b = example["question_content"]
        label = int(example['topic'])
        guid = "{}".format(example["id"])
        return InputExample(guid = guid, text_a=text_a, text_b=text_b, label=label)











PROCESSORS = {
    "super_glue.multirc": SuperglueMultiRCProcessor,
    "super_glue.boolq": SuperglueBoolQProcessor,
    "super_glue.cb": SuperglueCBProcessor,
    "super_glue.copa": SuperglueCOPAProcessor,
    "super_glue.rte": SuperglueRTEProcessor,
    "super_glue.wic": SuperglueWiCProcessor,
    "super_glue.wsc": SuperglueWSCProcessor,
    "super_glue.record": SuperglueRecordProcessor,
    "yahoo_answers_topics": YahooAnswersTopicsProcessor
}



from yacs.config import CfgNode
from .typing_dataset import PROCESSORS as TYPING_PROCESSORS
from .text_classification_dataset import PROCESSORS as TC_PROCESSORS
from .fewglue_dataset import PROCESSORS as SUPERGLUE_PROCESSORS
from .relation_classification_dataset import PROCESSORS as RC_PROCESSORS
from .lama_dataset import PROCESSORS as LAMA_PROCESSORS
from .conditional_generation_dataset import PROCESSORS as CG_PROCESSORS
from .utils import InputExample, InputFeatures
from .data_sampler import FewShotSampler
# support loading transformers datasets from https://huggingface.co/docs/datasets/
from .nli_dataset import PROCESSORS as NLI_PROCESSORS

from SohuPrompt.utils.logging import logger
from SohuPrompt.data_utils.huggingface_dataset import PROCESSORS as HF_PROCESSORS

PROCESSORS = {
    **TYPING_PROCESSORS,
    **TC_PROCESSORS,
    **SUPERGLUE_PROCESSORS,
    **RC_PROCESSORS,
    **LAMA_PROCESSORS,
    **CG_PROCESSORS,
    **LAMA_PROCESSORS,
    **HF_PROCESSORS,
    **NLI_PROCESSORS,
}


def load_dataset(config: CfgNode, return_class=True, test=False):
    r"""A plm loader using a global config.
    It will load the train, valid, and test set (if exists) simulatenously.

    Args:
        config (:obj:`CfgNode`): The global config from the CfgNode.
        return_class (:obj:`bool`): Whether return the data processor class
                    for future usage.

    Returns:
        :obj:`Optional[List[InputExample]]`: The train dataset.
        :obj:`Optional[List[InputExample]]`: The valid dataset.
        :obj:`Optional[List[InputExample]]`: The test dataset.
        :obj:"
    """
    dataset_config = config.dataset

    processor = PROCESSORS[dataset_config.name.lower()]()

    train_dataset = None
    valid_dataset = None
    if not test:
        try:
            train_dataset = processor.get_train_examples(dataset_config.path)
        except FileNotFoundError:
            logger.warning(f"Has no training dataset in {dataset_config.path}.")
        try:
            valid_dataset = processor.get_dev_examples(dataset_config.path)
        except FileNotFoundError:
            logger.warning(f"Has no validation dataset in {dataset_config.path}.")

    test_dataset = None
    try:
        test_dataset = processor.get_test_examples(dataset_config.path)
    except FileNotFoundError:
        logger.warning(f"Has no test dataset in {dataset_config.path}.")
    # checking whether downloaded.
    if (train_dataset is None) and \
       (valid_dataset is None) and \
       (test_dataset is None):
        logger.error("Dataset is empty. Either there is no download or the path is wrong. "+ \
        "If not downloaded, please `cd datasets/` and `bash download_xxx.sh`")
        exit()
    if return_class:
        return train_dataset, valid_dataset, test_dataset, processor
    else:
        return  train_dataset, valid_dataset, test_dataset




# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import List
from SohuPrompt.data_utils import InputExample
from SohuPrompt.data_utils.data_processor import DataProcessor

# logger = log.get_logger(__name__)



class SNLIProcessor(DataProcessor):
    """
    `The Stanford Natural Language Inference (SNLI) corpus <https://nlp.stanford.edu/projects/snli/>`_ is a dataset for natural language inference. It is first released in `A large annotated corpus for learning natural language inference (Bowman et al. 2015) <https://nlp.stanford.edu/pubs/snli_paper.pdf>`_

    We use the data released in `Making Pre-trained Language Models Better Few-shot Learners (Gao et al. 2020) <https://arxiv.org/pdf/2012.15723.pdf>`_

    Examples:

    ..  code-block:: python

        from SohuPrompt.data_utils.lmbff_dataset import PROCESSORS

        base_path = "datasets"

        dataset_name = "SNLI"
        dataset_path = os.path.join(base_path, dataset_name, '16-13')
        processor = PROCESSORS[dataset_name.lower()]()
        train_dataset = processor.get_train_examples(dataset_path)
        dev_dataset = processor.get_dev_examples(dataset_path)
        test_dataset = processor.get_test_examples(dataset_path)

        assert processor.get_num_labels() == 3
        assert processor.get_labels() == ['entailment', 'neutral', 'contradiction']
        assert len(train_dataset) == 549367
        assert len(dev_dataset) == 9842
        assert len(test_dataset) == 9824
        assert train_dataset[0].text_a == 'A person on a horse jumps over a broken down airplane.'
        assert train_dataset[0].text_b == 'A person is training his horse for a competition.'
        assert train_dataset[0].label == 1
    """
    def __init__(self):
        super().__init__()
        self.labels = ['entailment', 'neutral', 'contradiction']

    def get_examples(self, data_dir, split):
        path = os.path.join(data_dir, f"{split}.tsv")
        examples = []
        with open(path, encoding='utf-8')as f:
            lines = f.readlines()
            for idx, line in enumerate(lines[1:]):
                linelist = line.strip().split('\t')
                guid = "%s-%s" % (split, idx)
                label = linelist[-1]
                text_a = linelist[7]
                text_b = linelist[8]
                example = InputExample(guid=guid, text_a=text_a, text_b=text_b, label=self.get_label_id(label))
                examples.append(example)
        return examples

PROCESSORS = {
    "snli": SNLIProcessor
}


from functools import partial
import json
from SohuPrompt.data_utils.utils import InputExample
from transformers.tokenization_utils import PreTrainedTokenizer
from yacs.config import CfgNode
from SohuPrompt.data_utils import InputFeatures
import re
from SohuPrompt import Verbalizer
from typing import *
import torch
import torch.nn as nn
import torch.nn.functional as F
from SohuPrompt.utils.logging import logger
from SohuPrompt.utils.crossfit_metrics import *
import re


class GenerationVerbalizer(Verbalizer):
    r"""
    This verbalizer is useful when the label prediction is better defined by a piece of input.
    For example, in correference resolution, the tgt_text is a proper noun mentioned in the text.
    There is no fixed mapping between a class label and its label words. This verbalizer
    can be used as verbalizer of ``COPA`` and ``WSC`` datasets in SuperGlue.

    This verbalizer is especially powerful when combined with
    `All NLP Tasks Are Generation Tasks <https://arxiv.org/abs/2103.10360>`_ Paradigm (Also see
    `Crossfit <https://arxiv.org/abs/2104.08835>`_). It can make any piece of text the tgt_text. The tgt_text will then be filled in the `{"mask"}`.

    For example, when label word is ``"good"``, the tgt_text is ``"good"``;

    when label word is ``{"text":"good"}``, the tgt_text is also ``"good"``;

    when label word is ``{"meta":"choice1"}``, the tgt_text is the ``"meta['choice1']"`` field of the ``InputExample``;

    when label word is ``{"meta":"choice1"} {"placeholder", "text_a"} .``, the tgt_text is the ``"meta['choice1']"`` field of the ``InputExample``,
    followed by ``text_a`` field of the ``InputExample``, and then a ``'.'``;

    A use case can be seen in `Tutorial 4.1 <https://github.com/thunlp/OpenPrompt/blob/main/tutorial/4.1_all_tasks_are_generation.py>`_

    Args:
        tokenizer (:obj:`PreTrainedTokenizer`): The tokenizer of the current pre-trained model to point out the vocabulary.
        classes (:obj:`List[Any]`): The classes (or labels) of the current task.
        prefix (:obj:`str`, optional): The prefix string of the verbalizer (used in PLMs like RoBERTa, which is sensitive to prefix space)
        is_rule (:obj:`bool`, optional): When the verbalizer use the rule syntax of MixTemplate.
        label_words (:obj:`dict`, optional): The label words of the generation verbalizer

    Example:
    To use this verbalizer to train the T5 model to predict answer and explanation using two masks.

    When the template (Defined by :obj:`MixedTemplate`) is:
    >>> input_example = InputExample(text_a = "Can fish run?", meta={"answer":"no", "explanation": "The fish have no legs"}, label=0)
    >>> template = "{'placeholder':'text_a'} answer: {'mask'} explanation: {'mask'}"

    The verbalizer can be:
    >>> label_words = {0:["no", "{'meta':'explanation'}"], 1:["yes", "{'meta':'explanation'}"]}
    >>> verbalizer = GenerationVerbalizer(tokenizer, classes=None, is_rule=True, label_words=label_words)




    """
    def __init__(self,
                 tokenizer: PreTrainedTokenizer,
                 classes: Optional[List[str]] = None,
                 num_classes: Optional[int] = None,
                 is_rule: Optional[bool] = False,
                 label_words: Optional[dict] = None,
                ):
        if classes is None and label_words is not None:
            classes = list(label_words.keys())
        super().__init__(tokenizer=tokenizer, num_classes=num_classes, classes=classes)
        self.prefix = ''
        self.is_rule = is_rule
        self.mixed_token_start = "{"
        self.mixed_token_end = "}"

        if label_words is not None: # use label words as an initialization
            self.label_words = label_words

    def wrap_one_example(self, example: InputExample) -> List[Dict]:
        r"""Take an InputExample, and fill the tgt_text with label words
        """
        if not isinstance(self.label_words[example.label], list):
            label_word = [self.label_words[example.label]]
        else:
            label_word = self.label_words[example.label]

        if example.tgt_text is not None:
            logger.warning(f"The example already has tgt_text {example.tgt_text}, and will be filled with new label words, is this intended?")

        if not self.is_rule:
            instance_label_word =  label_word
        else:
            instance_label_word = [i(example) for i in label_word]  #(example)
        if len(instance_label_word) == 1:
            example.tgt_text = instance_label_word[0]
        else:
            example.tgt_text = instance_label_word
        return example


    def on_label_words_set(self):
        r"""
        Process the text into the label words (sometimes a function) according to the syntax of MixedTemplate
        """
        if isinstance(self.label_words[0], list):
            self.label_words = [x[0] for x in self.label_words]

        if self.is_rule:
            for id, label_word in enumerate(self.label_words):
                try:
                    d = self.parse_text(label_word)
                except:
                    raise RuntimeError(f"is_rule={self.is_rule} but label_word: {label_word} can't be converted to object.")
                self.label_words[id] = partial(lambda x, text: self.incorporate_text_example(text, x), text=d)



    def parse_text(self, text: str) -> List[Dict]:
        parsed = []
        i = 0
        while i < len(text):
            d = {"add_prefix_space": ' ' if (i > 0 and text[i-1] == ' ') else ''}
            while i < len(text) and text[i] == ' ':
                d["add_prefix_space"] = ''
                i = i + 1
            if i == len(text): break

            if text[i] != self.mixed_token_start:
                j = i + 1
                while j < len(text):
                    if text[j] == self.mixed_token_start:
                        break
                    j = j + 1
                d["text"] = text[i:j].rstrip(' ')
                i = j

            else:
                j = i + 1
                while j < len(text):
                    if text[j] == self.mixed_token_end:
                        break
                    j = j + 1
                if j == len(text):
                    raise ValueError(f"mixed_token_start {self.mixed_token_start} at position {i} has no corresponding mixed_token_end {self.mixed_token_end}")
                dict_str = '{'+text[i+1:j]+'}'
                try:
                    val = eval(dict_str)
                    if isinstance(val, set):
                        val = {k: None for k in val}
                    d.update(val)
                except:
                    import traceback
                    print(traceback.format_exc())
                    print(f"syntax error in {dict_str}")
                    exit()
                i = j + 1

            parsed.append(d)

        return parsed

    def incorporate_text_example(self,
                                 text,
                                 example: InputExample
                                ):
        text = text.copy()
        for i, d in enumerate(text):
            if 'placeholder' in d:
                text[i] = d["add_prefix_space"] + d.get("post_processing", lambda x:x)(getattr(example, d['placeholder']))
            elif 'meta' in d:
                text[i] = d["add_prefix_space"] + d.get("post_processing", lambda x:x)(example.meta[d['meta']])
            elif 'soft' in d:
                raise RuntimeError("soft token not supported in verbalizer") # unused
            elif 'mask' in d:
                raise RuntimeError("mask token not supported in verbalizer")
            elif 'special' in d:
                raise RuntimeError("special token not supported in verbalizer")
            elif 'text' in d:
                text[i] = d["add_prefix_space"] + d['text']
            else:
                raise ValueError(f'can not parse {d}')
        text = " ".join(text)
        return text











from yacs.config import CfgNode

def get_default_config():
    # OpenPrompt's default configuration options
    cfg = CfgNode(new_allowed=True)

    # ENVIRONMENT
    ###################################
    cfg.environment = CfgNode(new_allowed=True)
    cfg.environment.num_gpus = 1 # number of gpus to use
    cfg.environment.cuda_visible_devices = [0] # which index of cuda devices is visible to the program
    cfg.environment.local_rank = 0 # the main device in the cuda visible devices that your DataParallel model will put the model on.
                    # The following should holds: local_rank < len(cuda_visible_devices)
    cfg.environment.model_parallel = False  # whether to perform model parallel
    cfg.environment.device_map = None  # the device_map, such as "{0: [0, 1, 2], 1: [3, 4, 5, 6, 7, 8, 9], 2: [10, 11, 12, 13, 14, 15, 16],3: [17, 18, 19, 20, 21, 22, 23]}

    cfg.reproduce = CfgNode(new_allowed=True) # seed for reproduction
    #cfg.reproduce.seed = 100  # a seed for all everything

    # ONNX 
    ##################################
    cfg.onnx = CfgNode(new_allowed=True)
    cfg.onnx.infer = False

    # PLM PARAMETERS
    ##################################
    cfg.plm = CfgNode(new_allowed=True)
    cfg.plm.model_name = None # the model name, e.g. bert, roberta, gpt2, ...
                # for all the available model, please check the ./plms directory.
    cfg.plm.model_path = None
    # cfg.plm.specials_to_add = ['<pad>'] # always need to add pad token, if the tokenizer doesn't have one.
    cfg.plm.specials_to_add = None 
    cfg.plm.optimize = CfgNode(new_allowed=True)
    cfg.plm.optimize.name = 'AdamW'  # TODO: currently not in use.
    cfg.plm.optimize.freeze_para = False
    cfg.plm.optimize.no_decay = ['bias', 'LayerNorm.weight']
    cfg.plm.optimize.lr = 0.0005
    cfg.plm.optimize.weight_decay = 0.01
    cfg.plm.optimize.betas = [0.9, 0.999]
    cfg.plm.optimize.eps = 1.0e-8
    cfg.plm.optimize.scheduler = CfgNode(new_allowed=True)
    cfg.plm.optimize.scheduler.type = None      # by default, it will choose get_linear_schedule_with_warmup
    cfg.plm.optimize.scheduler.num_warmup_steps = 500

    ## LOGIN and CHECKPOINTING ##################################################
    ## in logging, each experiment will create a separate folder for saving log.txt
    ## , (full) config.json, and the checkpoint (if use the same path).
    ## logging is in the following format：
    ## ./log
    ##  - DIR_NAME_1
    ##    - log.txt
    ##    - config.yaml
    ##    - checkpoint.pt
    ##    - ...
    ##  - DIR_NAME_2
    ##    - ...
    ##
    cfg.logging = CfgNode(new_allowed=True)
    cfg.logging.path_base = 'logs' # the path base of all the logs.
    cfg.logging.file_level = 'NOTSET' # make sure it's an option of logging package
    cfg.logging.console_level = 'INFO' # make sure it's an option of logging package
    cfg.logging.unique_string = None  # the generated (or usr defined) unique string for one experiment.
    cfg.logging.unique_string_keys = ['dataset.name', 'plm.model_path', 'template', 'verbalizer', 'datetime'] # used to generate the unique string for saving
        #- dataset.name
        #- plm.model_path # only keep the last folder name in code,
            # .i.e ../.cache/roberta-large/ will save as roberta-large
        #- template
        #- verbalizer
        #- datetime  # a 12-digit string recording the date time of running the experiment, i.e., YYMMDDHHMMSS.
    cfg.logging.datetime_format = "%m%d%H%M%S%f" # only useful when unique_string_keys includes `datetime`.
        #  make sure it's a valid format for datetime package.
    cfg.logging.path = None # always keep none to let the config generate a full path according to
            # path_base and unique_string.
    cfg.logging.overwrite = True # if a same log path exists, overwrite it.

    # CHECKPOINT
    ######################################
    cfg.checkpoint = CfgNode(new_allowed=True) # checkpoint use the same directory as logging.
    cfg.checkpoint.save_latest = True # Normally set to False to reduce memory use, set
                        # to true to allow resuming learning process.
    cfg.checkpoint.save_best = True   # Keep saving the epoch of the best-performance.
    cfg.checkpoint.higher_better = True # is the metric to determine best checkpoint higher better?


    ## PIPELINE #######################################################

    cfg.train = CfgNode(new_allowed=True)
    cfg.train.num_epochs = 5 # the number of training epochs.
    cfg.train.num_training_steps = None
    cfg.train.batch_size = 2 # the batch_size.
    cfg.train.shuffle_data = True # whether shuffle the training data.
    cfg.train.teacher_forcing = False # whether perform teacher forcing in training.
                        # if true, the desired prediction on each mask will
                        # be filled in the mask.
    cfg.train.gradient_accumulation_steps = 1 # update weight  every N step of training.
                        # set 1 to disable gradient accumulation.
    cfg.train.max_grad_norm = -1.0 # <0 for unlimited gradients norm
    cfg.train.clean = False # set to True for not saving checkpoint and no tensorboard logging

    cfg.dev = CfgNode(new_allowed=True)
    cfg.dev.batch_size = 2 # evaluationn batch_size, can be a bit larger than training batch_size
    cfg.dev.shuffle_data = False # whether to perform data shuffling in evaluation

    cfg.test = CfgNode(new_allowed=True)
    cfg.test.batch_size = 2 # evaluationn batch_size, can be a bit larger than training batch_size
    cfg.test.shuffle_data = False # whether to perform data shuffling in evaluation

    ## TASK ##########################################################@
    cfg.task = 'classification'

    cfg.classification = CfgNode(new_allowed=True)
    cfg.classification.parent_config = 'task'
    cfg.classification.metric = ['micro-f1']  # the first one will be the main  to determine checkpoint. whether the higher metric value is better.
    cfg.classification.loss_function = 'cross_entropy' # the loss function for classification

    # LMBFF-classification config ###########################################################W
    cfg.classification.auto_t = False
    cfg.classification.auto_v = False

    cfg.template_generator = CfgNode(new_allowed=True)
    cfg.template_generator.plm = CfgNode(new_allowed=True)
    cfg.template_generator.plm.model_name = 't5' # the model name, e.g. bert, roberta, gpt2, ...
                # for all the available model, please check the ./plms directory.
    cfg.template_generator.plm.model_path = None
    cfg.template_generator.plm.specials_to_add = ['<pad>'] # always need to add pad token, if the tokenizer doesn't have one.

    cfg.template_generator.max_length = 20 # maximum length of generated template
    cfg.template_generator.target_number = 2 # number of parts to generate, e.g. in T5, every <extra_id_{}> token is one part
    cfg.template_generator.beam_width = 5
    cfg.template_generator.length_limit = None # List[str] length limit for each part of content
    cfg.template_generator.template = CfgNode(new_allowed=True)
    cfg.template_generator.template.text = None
    cfg.template_generator.template.mask_token = '<mask>'
    cfg.template_generator.template.placeholder_mapping = CfgNode(new_allowed=True)
    cfg.template_generator.template.placeholder_mapping['<text_a>'] = 'text_a'
    cfg.template_generator.template.placeholder_mapping['<text_b>'] = 'text_b'
    cfg.template_generator.template.file_path = None
    cfg.template_generator.template.choice = 0

    # verbalizer_generator, refer to https://arxiv.org/abs/2010.13641
    cfg.verbalizer_generator = CfgNode(new_allowed=True)
    cfg.verbalizer_generator.candidate_num = 1 # the number of candidates for further selection
    cfg.verbalizer_generator.label_word_num_per_class = 1
    cfg.verbalizer_generator.score_fct = 'llr' # the scoring function of label words selection. ``llr'' means log likelihood ratio, corresponding to Equation (7); ``ce'' means cross entropy, corresponding to Equation (6). As the paper points out, ``llr'' is significantly better than 'ce', we only keep it to matchthe original code.
    cfg.verbalizer_generator.normalize = True # whether to perform normalization of unbalanced training dataset, as Equation (5)



    cfg.generation = CfgNode(new_allowed=True) # Adding any arguments for generation here.
    cfg.generation.parent_config = 'task'
    cfg.generation.metric = ['sentence_bleu']
    #cfg.generation.max_length = 512   # the max_length of the generated sentence. INCLUDING the input_ids. So: generation.max_length > dataloader.max_seq_length
    cfg.generation.max_new_tokens = None
    cfg.generation.min_length = 5
    cfg.generation.temperature = 1.0
    cfg.generation.do_sample = False
    cfg.generation.top_k = 0
    cfg.generation.top_p = 0.9
    cfg.generation.repetition_penalty = 1.0 ##args.repetition_penalty,
    cfg.generation.num_beams = 5
    #cfg.generation.bad_words_ids = [[628, 198]]


    cfg.relation_classification = CfgNode(new_allowed=True)
    cfg.relation_classification.parent_config = 'task'

    ## DATASET #########################################################
    cfg.dataset = CfgNode(new_allowed=True)
    cfg.dataset.name = None   # the name of the dataset, for the supported choices,
            # please see the processors in ./data_utils/
    cfg.dataset.path = None  # whether is the dataset saved in your local machine.
    cfg.dataset.label_path_sep = None # label path separation token, only for hierarchical label

    ## DATALOADER ######################################################
    cfg.dataloader = CfgNode(new_allowed=True)
    cfg.dataloader.max_seq_length = 256 # max_seq_length
    cfg.dataloader.decoder_max_length = 256 # the decoder max length to truncate decoder input sequence
                        # if it is an encoder-decoder architecture. Note that it's not equavalent
                        # to generation.max_length which is used merely in the generation phase.
    cfg.dataloader.truncate_method = "head" # choosing from balanced, head, tail

    ## LEARINING SETTING  ####################################################
    cfg.learning_setting = None   # selecting from "full", "zero-shot", "few-shot"

    cfg.zero_shot = CfgNode(new_allowed=True)
    cfg.zero_shot.parent_config = 'learning_setting'

    cfg.few_shot = CfgNode(new_allowed=True)
    cfg.few_shot.parent_config = 'learning_setting'
    cfg.few_shot.few_shot_sampling = None

    cfg.sampling_from_train = CfgNode(new_allowed=True)
    cfg.sampling_from_train.parent_config = 'few_shot_sampling'
    cfg.sampling_from_train.num_examples_per_label = 10
    cfg.sampling_from_train.also_sample_dev = True
    cfg.sampling_from_train.num_examples_per_label_dev = 10
    cfg.sampling_from_train.seed = [123]

    ## CALIBRATION ###########################################################
    cfg.calibrate = None # leave blank to use no calibrate

    cfg.contextualized_calibrate = CfgNode(new_allowed=True)
    cfg.contextualized_calibrate.parent_config = 'calibrate'
    cfg.contextualized_calibrate.num_example = None
    cfg.contextualized_calibrate.use_split = 'train'

    cfg.pmi_calibrate = CfgNode(new_allowed=True)
    cfg.pmi_calibrate.parent_config = 'calibrate'

    ## PROMPT SPECIFIC CONFIG ############################################
    cfg.template = None
    cfg.verbalizer = None

    cfg.manual_template = CfgNode(new_allowed=True)
    cfg.manual_template.parent_config = 'template'
    cfg.manual_template.text = None
    cfg.manual_template.mask_token = '<mask>'
    cfg.manual_template.placeholder_mapping = CfgNode(new_allowed=True)
    cfg.manual_template.placeholder_mapping['<text_a>'] = 'text_a'
    cfg.manual_template.placeholder_mapping['<text_b>'] = 'text_b'
    cfg.manual_template.file_path = None
    cfg.manual_template.choice = 0
    cfg.manual_template.optimize = None  # the parameters related to optimize the template


    cfg.automatic_verbalizer = CfgNode(new_allowed=True)
    cfg.automatic_verbalizer.parent_config = 'verbalizer'
    cfg.automatic_verbalizer.num_cadidates = 1000
    cfg.automatic_verbalizer.label_word_num_per_class = 1
    cfg.automatic_verbalizer.num_searches = 1
    cfg.automatic_verbalizer.score_fct = 'llr'
    cfg.automatic_verbalizer.balance = True
    cfg.automatic_verbalizer.optimize = CfgNode(new_allowed=True)
    cfg.automatic_verbalizer.optimize.level = 'epoch'
    cfg.automatic_verbalizer.num_classes = None
    cfg.automatic_verbalizer.init_using_split = 'valid'

    cfg.one2one_verbalizer = CfgNode(new_allowed=True)
    cfg.one2one_verbalizer.parent_config = 'verbalizer'
    cfg.one2one_verbalizer.label_words = None
    cfg.one2one_verbalizer.prefix = " "
    cfg.one2one_verbalizer.multi_token_handler = 'first'
    cfg.one2one_verbalizer.file_path = None
    cfg.one2one_verbalizer.choice = None
    cfg.one2one_verbalizer.num_classes = None
    cfg.one2one_verbalizer.optimize = None

    cfg.manual_verbalizer = CfgNode(new_allowed=True)
    cfg.manual_verbalizer.parent_config = 'verbalizer'
    cfg.manual_verbalizer.label_words = None
    cfg.manual_verbalizer.prefix = " "
    cfg.manual_verbalizer.multi_token_handler = 'first'
    cfg.manual_verbalizer.file_path = None
    cfg.manual_verbalizer.choice = None
    cfg.manual_verbalizer.num_classes = None
    cfg.manual_verbalizer.optimize = None

    cfg.prefix_tuning_template = CfgNode(new_allowed=True)
    cfg.prefix_tuning_template.parent_config = 'template'
    cfg.prefix_tuning_template.text = None
    cfg.prefix_tuning_template.mask_token = '<mask>'
    cfg.prefix_tuning_template.num_token = 5
    cfg.prefix_tuning_template.placeholder_mapping = CfgNode(new_allowed=True)
    cfg.prefix_tuning_template.placeholder_mapping['<text_a>'] = 'text_a'
    cfg.prefix_tuning_template.placeholder_mapping['<text_b>'] = 'text_b'
    cfg.prefix_tuning_template.prefix_dropout = 0.0
    cfg.prefix_tuning_template.mid_dim = 512
    cfg.prefix_tuning_template.optimize = CfgNode(new_allowed=True)
    cfg.prefix_tuning_template.optimize.name = 'AdamW'
    cfg.prefix_tuning_template.optimize.lr = 0.00005
    cfg.prefix_tuning_template.optimize.betas = [0.9, 0.999]
    cfg.prefix_tuning_template.optimize.adam_epsilon = 1.0e-8
    cfg.prefix_tuning_template.optimize.weight_decay = 0.0
    cfg.prefix_tuning_template.optimize.no_decay = ['bias', 'LayerNorm.weight']
    cfg.prefix_tuning_template.optimize.scheduler = CfgNode(new_allowed=True)
    cfg.prefix_tuning_template.optimize.scheduler.num_warmup_steps = 0

    cfg.mixed_template = CfgNode(new_allowed=True)
    cfg.mixed_template.parent_config = 'template'
    cfg.mixed_template.text = None
    cfg.mixed_template.mask_token = '<mask>'
    cfg.mixed_template.placeholder_mapping = CfgNode(new_allowed=True)
    cfg.mixed_template.placeholder_mapping['<text_a>'] = 'text_a'
    cfg.mixed_template.placeholder_mapping['<text_b>'] = 'text_b'
    cfg.mixed_template.optimize = CfgNode(new_allowed=True)
    cfg.mixed_template.optimize.name = 'AdamW'
    cfg.mixed_template.optimize.lr = 0.00005
    cfg.mixed_template.optimize.betas = [0.9, 0.999]
    cfg.mixed_template.optimize.adam_epsilon = 1.0e-8
    cfg.mixed_template.optimize.weight_decay = 0.0
    cfg.mixed_template.optimize.no_decay = ['bias', 'LayerNorm.weight']
    cfg.mixed_template.optimize.scheduler = CfgNode(new_allowed=True)
    cfg.mixed_template.optimize.scheduler.num_warmup_steps = 0


    return cfg

